import asyncio

import pytest

from .models import db, User, UserType

pytestmark = pytest.mark.asyncio


async def test(bind):
    await User.create(nickname='test')
    assert isinstance(await User.query.gino.first(), User)
    bind.update_execution_options(return_model=False)
    assert not isinstance(await User.query.gino.first(), User)
    async with db.acquire() as conn:
        assert isinstance(
            await conn.execution_options(return_model=True).first(User.query),
            User)
        assert not isinstance(await User.query.execution_options(
            return_model=False).gino.first(), User)
    assert isinstance(
        await User.query.execution_options(return_model=True).gino.first(),
        User)
    assert not isinstance(await User.query.gino.first(), User)
    bind.update_execution_options(return_model=True)
    assert isinstance(await User.query.gino.first(), User)


# noinspection PyProtectedMember
async def test_compiled_first_not_found(bind):
    async with bind.acquire() as conn:
        with pytest.raises(LookupError, match='No such execution option'):
            result = conn._execute('SELECT NOW()', (), {})
            result.context._compiled_first_opt('nonexist')


# noinspection PyUnusedLocal
async def test_query_ext(bind):
    q = User.query
    assert q.gino.query is q

    u = await User.create(nickname='test')
    assert isinstance(await User.query.gino.first(), User)

    row = await User.query.gino.return_model(False).first()
    assert not isinstance(row, User)
    assert row == (
        u.id, 'test', {'age': 18, 'birthday': '1970-01-01T00:00:00.000000'},
        UserType.USER, None)

    row = await User.query.gino.model(None).first()
    assert not isinstance(row, User)
    assert row == (
        u.id, 'test', {'age': 18, 'birthday': '1970-01-01T00:00:00.000000'},
        UserType.USER, None)

    row = await db.select([User.id, User.nickname, User.type]).gino.first()
    assert not isinstance(row, User)
    assert row == (u.id, 'test', UserType.USER)

    user = await db.select(
        [User.id, User.nickname, User.type]).gino.model(User).first()
    assert isinstance(user, User)
    assert user.id is not None
    assert user.nickname == 'test'
    assert user.type == UserType.USER

    with pytest.raises(asyncio.TimeoutError):
        await db.select([db.func.pg_sleep(1), User.id]).gino.timeout(
            0.1).status()



# noinspection PyPackageRequirements
from sanic.exceptions import NotFound
from sqlalchemy.engine.url import URL

from ..api import Gino as _Gino, GinoExecutor as _Executor
from ..engine import GinoConnection as _Connection, GinoEngine as _Engine
from ..strategies import GinoStrategy


class SanicModelMixin:
    @classmethod
    async def get_or_404(cls, *args, **kwargs):
        # noinspection PyUnresolvedReferences
        rv = await cls.get(*args, **kwargs)
        if rv is None:
            raise NotFound('{} is not found'.format(cls.__name__))
        return rv


# noinspection PyClassHasNoInit
class GinoExecutor(_Executor):
    async def first_or_404(self, *args, **kwargs):
        rv = await self.first(*args, **kwargs)
        if rv is None:
            raise NotFound('No such data')
        return rv


# noinspection PyClassHasNoInit
class GinoConnection(_Connection):
    async def first_or_404(self, *args, **kwargs):
        rv = await self.first(*args, **kwargs)
        if rv is None:
            raise NotFound('No such data')
        return rv


# noinspection PyClassHasNoInit
class GinoEngine(_Engine):
    connection_cls = GinoConnection

    async def first_or_404(self, *args, **kwargs):
        rv = await self.first(*args, **kwargs)
        if rv is None:
            raise NotFound('No such data')
        return rv


class SanicStrategy(GinoStrategy):
    name = 'sanic'
    engine_cls = GinoEngine


SanicStrategy()


# noinspection PyClassHasNoInit
class Gino(_Gino):
    """Support Sanic web server.

    By :meth:`init_app` GINO registers a few hooks on Sanic, so that GINO could
    use database configuration in Sanic ``config`` to initialize the bound
    engine.

    A lazy connection context is enabled by default for every request. You can
    change this default behavior by setting ``DB_USE_CONNECTION_FOR_REQUEST``
    config value to ``False``. By default, a database connection is borrowed on
    the first query, shared in the same execution context, and returned to the
    pool on response. If you need to release the connection early in the middle
    to do some long-running tasks, you can simply do this::

        await request['connection'].release(permanent=False)

    """
    model_base_classes = _Gino.model_base_classes + (SanicModelMixin,)
    query_executor = GinoExecutor

    def __init__(self, app=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        if app.config.setdefault('DB_USE_CONNECTION_FOR_REQUEST', True):
            @app.middleware('request')
            async def on_request(request):
                request['connection'] = await self.acquire(lazy=True)

            @app.middleware('response')
            async def on_response(request, _):
                conn = request.pop('connection', None)
                if conn is not None:
                    await conn.release()

        @app.listener('before_server_start')
        async def before_server_start(_, loop):
            if app.config.get('DB_DSN'):
                dsn = app.config.DB_DSN
            else:
                dsn = URL(
                    drivername=app.config.setdefault('DB_DRIVER', 'asyncpg'),
                    host=app.config.setdefault('DB_HOST', 'localhost'),
                    port=app.config.setdefault('DB_PORT', 5432),
                    username=app.config.setdefault('DB_USER', 'postgres'),
                    password=app.config.setdefault('DB_PASSWORD', ''),
                    database=app.config.setdefault('DB_DATABASE', 'postgres'),
                )

            await self.set_bind(
                dsn,
                echo=app.config.setdefault('DB_ECHO', False),
                min_size=app.config.setdefault('DB_POOL_MIN_SIZE', 5),
                max_size=app.config.setdefault('DB_POOL_MAX_SIZE', 10),
                ssl=app.config.setdefault('DB_SSL'),
                loop=loop,
            )

        @app.listener('after_server_stop')
        async def after_server_stop(_, loop):
            await self.pop_bind().close()

    async def first_or_404(self, *args, **kwargs):
        rv = await self.first(*args, **kwargs)
        if rv is None:
            raise NotFound('No such data')
        return rv

    async def set_bind(self, bind, loop=None, **kwargs):
        kwargs.setdefault('strategy', 'sanic')
        return await super().set_bind(bind, loop=loop, **kwargs)



from gino import UninitializedError
import pytest

from .models import db, User

pytestmark = pytest.mark.asyncio


@pytest.fixture
def names(sa_engine):
    rv = {'11', '22', '33'}
    sa_engine.execute(User.__table__.insert(),
                      [dict(name=name) for name in rv])
    yield rv
    sa_engine.execute('DELETE FROM gino_users')


# noinspection PyUnusedLocal,PyShadowingNames
async def test_bind(bind, names):
    with pytest.raises(ValueError, match='No Connection in context'):
        async for u in User.query.gino.iterate():
            assert False, 'Should not reach here'
    with pytest.raises(ValueError, match='No Connection in context'):
        await User.query.gino.iterate()
    with pytest.raises(ValueError, match='No Connection in context'):
        await db.iterate(User.query)

    result = set()
    async with bind.transaction():
        async for u in User.query.gino.iterate():
            result.add(u.nickname)
    assert names == result

    result = set()
    async with bind.transaction():
        async for u in db.iterate(User.query):
            result.add(u.nickname)
    assert names == result

    result = set()
    async with bind.transaction():
        cursor = await User.query.gino.iterate()
        result.add((await cursor.next()).nickname)
        assert names != result
        result.update([u.nickname for u in await cursor.many(1)])
        assert names != result
        result.update([u.nickname for u in await cursor.many(2)])
        assert names == result
        result.update([u.nickname for u in await cursor.many(2)])
        assert names == result
        assert await cursor.next() is None

    with pytest.raises(ValueError, match='too many multiparams'):
        async with bind.transaction():
            await db.iterate(User.insert().returning(User.nickname), [
                dict(nickname='444'),
                dict(nickname='555'),
                dict(nickname='666'),
            ])

    result = set()
    async with bind.transaction():
        cursor = await User.query.gino.iterate()
        await cursor.forward(1)
        result.add((await cursor.next()).nickname)
        assert names != result
        result.update([u.nickname for u in await cursor.many(1)])
        assert names != result
        result.update([u.nickname for u in await cursor.many(2)])
        assert names != result
        assert await cursor.next() is None


# noinspection PyUnusedLocal,PyShadowingNames
async def test_basic(engine, names):
    result = set()
    async with engine.transaction() as tx:
        with pytest.raises(UninitializedError):
            await db.iterate(User.query)
        result = set()
        async for u in tx.connection.iterate(User.query):
            result.add(u.nickname)
        async for u in tx.connection.execution_options(
                timeout=1).iterate(User.query):
            result.add(u.nickname)
        assert names == result

        result = set()
        cursor = await tx.connection.iterate(User.query)
        result.update([u.nickname for u in await cursor.many(2)])
        assert names != result
        result.update([u.nickname for u in await cursor.many(2)])
        assert names == result



import ssl

import asyncpg
import pytest
import sqlalchemy
from async_generator import yield_, async_generator

import gino
from .models import db, DB_ARGS, PG_URL, random_name

ECHO = False


@pytest.fixture(scope='module')
def sa_engine():
    rv = sqlalchemy.create_engine(PG_URL, echo=ECHO)
    db.create_all(rv)
    yield rv
    db.drop_all(rv)
    rv.dispose()


@pytest.fixture
@async_generator
async def engine(sa_engine):
    e = await gino.create_engine(PG_URL, echo=ECHO)
    await yield_(e)
    await e.close()
    sa_engine.execute('DELETE FROM gino_user_settings')
    sa_engine.execute('DELETE FROM gino_users')


# noinspection PyUnusedLocal,PyShadowingNames
@pytest.fixture
@async_generator
async def bind(sa_engine):
    async with db.with_bind(PG_URL, echo=ECHO) as e:
        await yield_(e)
    sa_engine.execute('DELETE FROM gino_user_settings')
    sa_engine.execute('DELETE FROM gino_users')


# noinspection PyUnusedLocal,PyShadowingNames
@pytest.fixture
@async_generator
async def asyncpg_pool(sa_engine):
    async with asyncpg.create_pool(**DB_ARGS) as rv:
        await yield_(rv)
        await rv.execute('DELETE FROM gino_user_settings')
        await rv.execute('DELETE FROM gino_users')


@pytest.fixture
def ssl_ctx():
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    return ctx



import random

from gino.exceptions import UninitializedError
import pytest

from .models import db, PG_URL, User

pytestmark = pytest.mark.asyncio


# noinspection PyUnusedLocal
async def test_create(bind):
    nickname = 'test_create_{}'.format(random.random())
    u = await User.create(nickname=nickname)
    assert u.id is not None
    assert u.nickname == nickname
    return u


async def test_get(bind):
    u1 = await test_create(bind)
    u2 = await User.get(u1.id)
    assert u1.id == u2.id
    assert u1.nickname == u2.nickname
    assert u1 is not u2


# noinspection PyUnusedLocal
async def test_unbind(asyncpg_pool):
    await db.set_bind(PG_URL)
    await test_create(None)
    await db.pop_bind().close()
    db.bind = None
    with pytest.raises(UninitializedError):
        await test_create(None)
    # test proper exception when engine is not initialized
    with pytest.raises(UninitializedError):
        db.bind.first = lambda x: 1


async def test_db_api(bind, random_name):
    assert await db.scalar(
        User.insert().values(name=random_name).returning(
            User.nickname)) == random_name
    assert (await db.first(User.query.where(
        User.nickname == random_name))).nickname == random_name
    assert len(
        await db.all(User.query.where(User.nickname == random_name))) == 1
    assert (await db.status(User.delete.where(
        User.nickname == random_name)))[0] == 'DELETE 1'
    stmt, params = db.compile(User.query.where(User.id == 3))
    assert params[0] == 3



from .api import Gino  # NOQA
from .engine import GinoEngine, GinoConnection  # NOQA
from .exceptions import *  # NOQA
from .strategies import GinoStrategy  # NOQA


def create_engine(*args, **kwargs):
    from sqlalchemy import create_engine

    kwargs.setdefault('strategy', 'gino')
    return create_engine(*args, **kwargs)


__version__ = '0.8.1'



from datetime import datetime

import pytest
from .models import db, User

pytestmark = pytest.mark.asyncio


async def test_compiled_and_bindparam(bind):
    async with db.acquire() as conn:
        # noinspection PyArgumentList
        ins = await conn.prepare(User.insert().returning(
            *User).execution_options(loader=User))
        users = {}
        for name in '12345':
            u = await ins.first(name=name)
            assert u.nickname == name
            users[u.id] = u
        get = await conn.prepare(
            User.query.where(User.id == db.bindparam('uid')))
        for key in users:
            u = await get.first(uid=key)
            assert u.nickname == users[key].nickname
            assert (await get.all(uid=key))[0].nickname == u.nickname

        assert await get.scalar(uid=-1) is None

        with pytest.raises(ValueError, match='does not support multiple'):
            await get.all([dict(uid=1), dict(uid=2)])

        delete = await conn.prepare(
            User.delete.where(User.nickname == db.bindparam('name')))
        for name in '12345':
            msg = await delete.status(name=name)
            assert msg == 'DELETE 1'


async def test_statement(engine):
    async with engine.acquire() as conn:
        stmt = await conn.prepare('SELECT now()')
        last = None
        for i in range(5):
            now = await stmt.scalar()
            assert isinstance(now, datetime)
            assert last != now
            last = now



import pytest

from .models import PG_URL, User

pytestmark = pytest.mark.asyncio


async def test_anonymous(sa_engine):
    import gino
    e = await gino.create_engine(PG_URL, statement_cache_size=0)
    async with e.acquire() as conn:
        # noinspection PyProtectedMember
        assert conn.raw_connection._stmt_cache.get_max_size() == 0
        await conn.first(User.query.where(User.id == 1))
        # anonymous statement should not be closed
        await conn.first(User.query.where(User.id == 1))
    await e.close()



import pytest

from .models import db, User, qsize

pytestmark = pytest.mark.asyncio


async def _init(bind):
    from .test_crud import test_create
    u = await test_create(bind)

    def get_name():
        return User.select('nickname').where(User.id == u.id).gino.scalar()

    return u, get_name


async def test_connection_ctx(bind, mocker):
    init_size = qsize(bind)
    u, get_name = await _init(bind)

    assert await get_name() != 'commit'

    async with bind.acquire() as conn:
        async with conn.transaction():
            await u.update(nickname='commit').apply()
    assert await get_name() == 'commit'

    with pytest.raises(ZeroDivisionError):
        async with bind.acquire() as conn:
            async with conn.transaction():
                await u.update(nickname='rollback').apply()
                assert await get_name() == 'rollback'
                raise ZeroDivisionError

    assert await get_name() == 'commit'

    async with bind.acquire() as conn:
        tx = await conn.transaction().__aenter__()
        await u.update(nickname='rollback').apply()
        assert await get_name() == 'rollback'
        mocker.patch(
            'asyncpg.transaction.Transaction.commit').side_effect = IndexError
        with pytest.raises(IndexError):
            await tx.__aexit__(None, None, None)
        # clean up, and to simulate commit failed
        mocker.stopall()
        await tx._tx.rollback()
        assert await get_name() == 'commit'
    assert await get_name() == 'commit'

    assert init_size == qsize(bind)


async def test_connection_await(bind):
    init_size = qsize(bind)
    u, get_name = await _init(bind)

    assert await get_name() != 'commit'

    async with bind.acquire() as conn:
        tx = await conn.transaction()
        await u.update(nickname='commit').apply()
        await tx.commit()
    assert await get_name() == 'commit'

    async with bind.acquire() as conn:
        tx = await conn.transaction()
        await u.update(nickname='rollback').apply()
        assert await get_name() == 'rollback'
        await tx.rollback()

    assert await get_name() == 'commit'

    # Neither commit nor rollback, should rollback
    async with bind.acquire() as conn:
        await conn.transaction()
        await u.update(nickname='rollback').apply()
        assert await get_name() == 'rollback'

    assert await get_name() == 'commit'

    assert init_size == qsize(bind)


async def test_engine(bind):
    init_size = qsize(bind)
    u, get_name = await _init(bind)

    assert await get_name() != 'commit'

    async with bind.transaction():
        await u.update(nickname='commit').apply()
    assert await get_name() == 'commit'

    with pytest.raises(ZeroDivisionError):
        async with bind.transaction():
            await u.update(nickname='rollback').apply()
            raise ZeroDivisionError
    assert await get_name() == 'commit'
    assert init_size == qsize(bind)


async def test_begin_failed(bind, mocker):
    from asyncpg.transaction import Transaction
    init_size = qsize(bind)
    mocker.patch('asyncpg.transaction.Transaction.start')
    Transaction.start.side_effect = ZeroDivisionError
    with pytest.raises(ZeroDivisionError):
        async with bind.transaction():
            pass  # pragma: no cover
    assert init_size == qsize(bind)


async def test_commit_failed(bind, mocker):
    from asyncpg.transaction import Transaction
    init_size = qsize(bind)
    mocker.patch('asyncpg.transaction.Transaction._Transaction__commit')
    # noinspection PyUnresolvedReferences,PyProtectedMember
    Transaction._Transaction__commit.side_effect = ZeroDivisionError
    with pytest.raises(ZeroDivisionError):
        async with bind.transaction():
            pass
    assert init_size == qsize(bind)


async def test_reuse(bind):
    from asyncpg.transaction import Transaction
    init_size = qsize(bind)
    async with db.acquire() as conn:
        async with db.transaction() as tx:
            assert tx.connection.raw_connection is conn.raw_connection
            assert isinstance(tx.raw_transaction, Transaction)
            async with db.transaction() as tx2:
                assert tx2.connection.raw_connection is conn.raw_connection
            async with db.transaction(reuse=False) as tx2:
                assert tx2.connection.raw_connection is not conn.raw_connection
        async with db.transaction(reuse=False) as tx:
            assert tx.connection.raw_connection is not conn.raw_connection
            async with db.transaction() as tx2:
                assert (tx2.connection.raw_connection is
                        tx.connection.raw_connection)
            async with db.transaction(reuse=False) as tx2:
                assert tx2.connection.raw_connection is not conn.raw_connection
                assert (tx2.connection.raw_connection is not
                        tx.connection.raw_connection)
    with pytest.raises(ValueError, match='already released'):
        await conn.release()
    assert init_size == qsize(bind)


async def test_nested(bind):
    init_size = qsize(bind)
    u, get_name = await _init(bind)

    name = await get_name()
    assert u.nickname == name

    async with bind.transaction():
        await u.update(nickname='first').apply()
        async with bind.transaction():
            pass

    assert init_size == qsize(bind)


# noinspection PyUnreachableCode,PyUnusedLocal
async def test_early_end(bind):
    init_size = qsize(bind)
    u, get_name = await _init(bind)

    assert await get_name() != 'ininin'

    async with bind.transaction() as tx:
        async with bind.transaction():
            async with bind.transaction():
                await u.update(nickname='ininin').apply()
                tx.raise_commit()
                assert False, 'Should not reach here'
            assert False, 'Should not reach here'
        assert False, 'Should not reach here'

    assert await get_name() == 'ininin'
    assert init_size == qsize(bind)

    async with bind.transaction() as tx:
        async with bind.transaction():
            async with bind.transaction():
                await u.update(nickname='nonono').apply()
                assert await get_name() == 'nonono'
                tx.raise_rollback()
                assert False, 'Should not reach here'
            assert False, 'Should not reach here'
        assert False, 'Should not reach here'

    assert await get_name() == 'ininin'
    assert init_size == qsize(bind)

    reached = 0

    async with bind.transaction():
        async with bind.transaction() as tx:
            async with bind.transaction():
                await u.update(nickname='nonono').apply()
                assert await get_name() == 'nonono'
                tx.raise_rollback()
                assert False, 'Should not reach here'
            assert False, 'Should not reach here'
        reached += 1
        assert await get_name() == 'ininin'

    assert await get_name() == 'ininin'
    assert init_size == qsize(bind)
    assert reached == 1

    async with bind.transaction():
        async with bind.transaction() as tx:
            async with bind.transaction():
                await u.update(nickname='nonono').apply()
                assert await get_name() == 'nonono'
                tx.raise_commit()
                assert False, 'Should not reach here'
            assert False, 'Should not reach here'
        reached += 1
        assert await get_name() == 'nonono'

    assert await get_name() == 'nonono'
    assert init_size == qsize(bind)
    assert reached == 2


# noinspection PyUnreachableCode
async def test_end_raises_in_with(engine):
    async with engine.transaction() as tx:
        with pytest.raises(AssertionError, match='Illegal in managed mode'):
            await tx.commit()
        await tx.raise_commit()
        assert False, 'Should not reach here'

    async with engine.transaction() as tx:
        with pytest.raises(AssertionError, match='Illegal in managed mode'):
            await tx.rollback()
        await tx.raise_rollback()
        assert False, 'Should not reach here'


async def test_base_exception(engine):
    async with engine.transaction() as tx:
        # noinspection PyBroadException
        try:
            await tx.raise_commit()
        except Exception:
            assert False, 'Should not reach here'
        assert False, 'Should not reach here'


async def test_no_rollback_on_commit_fail(engine, mocker):
    mocker.patch(
        'asyncpg.transaction.Transaction.commit').side_effect = IndexError
    async with engine.acquire() as conn:
        tx = await conn.transaction().__aenter__()
        rollback = mocker.patch.object(tx._tx, 'rollback')
        with pytest.raises(IndexError):
            await tx.__aexit__(None, None, None)
        assert not rollback.called



import pytest
from sqlalchemy import Table, Column, Integer, String, MetaData, ForeignKey
from sqlalchemy.engine.result import RowProxy

from .models import PG_URL

pytestmark = pytest.mark.asyncio


async def test_engine_only():
    import gino
    from gino.schema import GinoSchemaVisitor

    metadata = MetaData()

    users = Table(
        'users', metadata,
        Column('id', Integer, primary_key=True),
        Column('name', String),
        Column('fullname', String),
    )

    Table(
        'addresses', metadata,
        Column('id', Integer, primary_key=True),
        Column('user_id', None, ForeignKey('users.id')),
        Column('email_address', String, nullable=False)
    )

    engine = await gino.create_engine(PG_URL)
    await GinoSchemaVisitor(metadata).create_all(engine)
    try:
        ins = users.insert().values(name='jack', fullname='Jack Jones')
        await engine.status(ins)
        res = await engine.all(users.select())
        assert isinstance(res[0], RowProxy)
    finally:
        await GinoSchemaVisitor(metadata).drop_all(engine)


async def test_core():
    from gino import Gino

    db = Gino()

    users = db.Table(
        'users', db,
        db.Column('id', db.Integer, primary_key=True),
        db.Column('name', db.String),
        db.Column('fullname', db.String),
    )

    db.Table(
        'addresses', db,
        db.Column('id', db.Integer, primary_key=True),
        db.Column('user_id', None, db.ForeignKey('users.id')),
        db.Column('email_address', db.String, nullable=False)
    )

    async with db.with_bind(PG_URL):
        await db.gino.create_all()
        try:
            await users.insert().values(
                name='jack',
                fullname='Jack Jones',
            ).gino.status()
            res = await users.select().gino.all()
            assert isinstance(res[0], RowProxy)
        finally:
            await db.gino.drop_all()


async def test_orm():
    from gino import Gino

    db = Gino()

    class User(db.Model):
        __tablename__ = 'users'

        id = db.Column(db.Integer, primary_key=True)
        name = db.Column(db.String)
        fullname = db.Column(db.String)

    class Address(db.Model):
        __tablename__ = 'addresses'

        id = db.Column(db.Integer, primary_key=True)
        user_id = db.Column(None, db.ForeignKey('users.id'))
        email_address = db.Column(db.String, nullable=False)

    async with db.with_bind(PG_URL):
        await db.gino.create_all()
        try:
            await User.create(name='jack', fullname='Jack Jones')
            res = await User.query.gino.all()
            assert isinstance(res[0], User)
        finally:
            await db.gino.drop_all()



import random
from datetime import datetime

from async_generator import yield_, async_generator
import pytest
from sqlalchemy import select
from sqlalchemy.sql.functions import count

from gino.loader import AliasLoader
from .models import db, User, Team, Company

pytestmark = pytest.mark.asyncio


@pytest.fixture
@async_generator
async def user(bind):
    c = await Company.create()
    t1 = await Team.create(company_id=c.id)
    t2 = await Team.create(company_id=c.id, parent_id=t1.id)
    u = await User.create(team_id=t2.id)
    u.team = t2
    t2.parent = t1
    t2.company = c
    t1.company = c
    await yield_(u)
    await User.delete.gino.status()
    await Team.delete.gino.status()
    await Company.delete.gino.status()


async def test_model_alternative(user):
    u = await User.query.gino.load(User).first()
    assert isinstance(u, User)
    assert u.id == user.id
    assert u.nickname == user.nickname


async def test_scalar(user):
    name = await User.query.gino.load(User.nickname).first()
    assert user.nickname == name

    uid, name = await User.query.gino.load((User.id, User.nickname)).first()
    assert user.id == uid
    assert user.nickname == name


async def test_model_load(user):
    u = await User.query.gino.load(User.load('nickname', User.team_id)).first()
    assert isinstance(u, User)
    assert u.id is None
    assert u.nickname == user.nickname
    assert u.team_id == user.team.id

    with pytest.raises(TypeError):
        await User.query.gino.load(User.load(123)).first()

    with pytest.raises(AttributeError):
        await User.query.gino.load(User.load(Team.id)).first()


async def test_216_model_load_passive_partial(user):
    u = await db.select([User.nickname]).gino.model(User).first()
    assert isinstance(u, User)
    assert u.id is None
    assert u.nickname == user.nickname


async def test_load_relationship(user):
    u = await User.outerjoin(Team).select().gino.load(
        User.load(team=Team)).first()
    assert isinstance(u, User)
    assert u.id == user.id
    assert u.nickname == user.nickname
    assert isinstance(u.team, Team)
    assert u.team.id == user.team.id
    assert u.team.name == user.team.name


async def test_load_nested(user):
    for u in (
        await User.outerjoin(Team).outerjoin(Company).select().gino.load(
            User.load(team=Team.load(company=Company))).first(),
        await User.load(team=Team.load(company=Company)).gino.first(),
        await User.load(team=Team.load(company=Company.on(
                Team.company_id == Company.id))).gino.first(),
        await User.load(team=Team.load(company=Company).on(
                User.team_id == Team.id)).gino.first(),
        await User.load(team=Team.on(User.team_id == Team.id).load(
            company=Company)).gino.first(),
    ):
        assert isinstance(u, User)
        assert u.id == user.id
        assert u.nickname == user.nickname
        assert isinstance(u.team, Team)
        assert u.team.id == user.team.id
        assert u.team.name == user.team.name
        assert isinstance(u.team.company, Company)
        assert u.team.company.id == user.team.company.id
        assert u.team.company.name == user.team.company.name


async def test_func(user):
    def loader(row, context):
        rv = User(id=row[User.id], nickname=row[User.nickname])
        rv.team = Team(id=row[Team.id], name=row[Team.name])
        rv.team.company = Company(id=row[Company.id], name=row[Company.name])
        return rv

    u = await User.outerjoin(Team).outerjoin(Company).select().gino.load(
        loader).first()
    assert isinstance(u, User)
    assert u.id == user.id
    assert u.nickname == user.nickname
    assert isinstance(u.team, Team)
    assert u.team.id == user.team.id
    assert u.team.name == user.team.name
    assert isinstance(u.team.company, Company)
    assert u.team.company.id == user.team.company.id
    assert u.team.company.name == user.team.company.name


async def test_adjacency_list(user):
    group = Team.alias()

    with pytest.raises(AttributeError):
        group.non_exist()

    # noinspection PyUnusedLocal
    def loader(row, context):
        rv = User(id=row[User.id], nickname=row[User.nickname])
        rv.team = Team(id=row[Team.id], name=row[Team.name])
        rv.team.parent = Team(id=row[group.id], name=row[group.name])
        return rv

    for exp in (loader,
                User.load(team=Team.load(parent=group)),
                User.load(team=Team.load(parent=group.load('id', 'name'))),
                User.load(team=Team.load(parent=group.load()))):
        u = await User.outerjoin(
            Team
        ).outerjoin(
            group, Team.parent_id == group.id
        ).select().gino.load(exp).first()

        assert isinstance(u, User)
        assert u.id == user.id
        assert u.nickname == user.nickname
        assert isinstance(u.team, Team)
        assert u.team.id == user.team.id
        assert u.team.name == user.team.name
        assert isinstance(u.team.parent, Team)
        assert u.team.parent.id == user.team.parent.id
        assert u.team.parent.name == user.team.parent.name


async def test_alias_loader_columns(user):
    user_alias = User.alias()
    base_query = user_alias.outerjoin(Team).select()

    query = base_query.execution_options(loader=AliasLoader(user_alias, 'id'))
    u = await query.gino.first()
    assert u.id is not None


async def test_multiple_models_in_one_query(bind):
    for _ in range(3):
        await User.create()

    ua1 = User.alias()
    ua2 = User.alias()
    join_query = select([ua1, ua2]).where(ua1.id < ua2.id)
    result = await join_query.gino.load((ua1.load('id'), ua2.load('id'))).all()
    assert len(result) == 3
    for u1, u2 in result:
        assert u1.id is not None
        assert u2.id is not None
        assert u1.id < u2.id


async def test_loader_with_aggregation(user):
    count_col = count().label('count')
    user_count = select(
        [User.team_id, count_col]
    ).group_by(
        User.team_id
    ).alias()
    query = Team.outerjoin(user_count).select()
    result = await query.gino.load(
        (Team.id, Team.name, user_count.columns.team_id, count_col)
    ).all()
    assert len(result) == 2
    # team 1 doesn't have users, team 2 has 1 user
    # third and forth columns are None for team 1
    for team_id, team_name, user_team_id, user_count in result:
        if team_id == user.team_id:
            assert team_name == user.team.name
            assert user_team_id == user.team_id
            assert user_count == 1
        else:
            assert team_id is not None
            assert team_name is not None
            assert user_team_id is None
            assert user_count is None


async def test_adjacency_list_query_builder(user):
    group = Team.alias()
    u = await User.load(team=Team.load(parent=group.on(
        Team.parent_id == group.id))).gino.first()

    assert isinstance(u, User)
    assert u.id == user.id
    assert u.nickname == user.nickname
    assert isinstance(u.team, Team)
    assert u.team.id == user.team.id
    assert u.team.name == user.team.name
    assert isinstance(u.team.parent, Team)
    assert u.team.parent.id == user.team.parent.id
    assert u.team.parent.name == user.team.parent.name


async def test_literal(user):
    sample = tuple(random.random() for _ in range(5))
    now = db.Column('time', db.DateTime())
    row = await db.first(db.text(
        'SELECT now() AT TIME ZONE \'UTC\''
    ).columns(now).gino.load(
        sample + (lambda r, c: datetime.utcnow(), now,)).query)
    assert row[:5] == sample
    assert isinstance(row[-2], datetime)
    assert isinstance(row[-1], datetime)
    assert row[-1] <= row[-2]


async def test_load_one_to_many(user):
    # noinspection PyListCreation
    uids = [user.id]
    uids.append((await User.create(nickname='1', team_id=user.team.id)).id)
    uids.append((await User.create(nickname='1', team_id=user.team.id)).id)
    uids.append((await User.create(nickname='2',
                                   team_id=user.team.parent.id)).id)
    query = User.outerjoin(Team).outerjoin(Company).select()
    companies = await query.gino.load(
        Company.distinct(Company.id).load(
            add_team=Team.load(add_member=User).distinct(Team.id))).all()
    assert len(companies) == 1
    company = companies[0]
    assert isinstance(company, Company)
    assert company.id == user.team.company_id
    assert company.name == user.team.company.name
    assert len(company.teams) == 2
    for team in company.teams:
        if team.id == user.team.id:
            assert len(team.members) == 3
            for u in team.members:
                if u.nickname == user.nickname:
                    assert isinstance(u, User)
                    assert u.id == user.id
                    uids.remove(u.id)
                if u.nickname in {'1', '2'}:
                    uids.remove(u.id)
        else:
            assert len(team.members) == 1
            uids.remove(list(team.members)[0].id)
    assert uids == []

    # test distinct many-to-one
    query = User.outerjoin(Team).select().where(Team.id == user.team.id)
    users = await query.gino.load(
        User.load(team=Team.distinct(Team.id))).all()
    assert len(users) == 3
    assert users[0].team is users[1].team
    assert users[0].team is users[2].team


async def test_distinct_none(bind):
    u = await User.create()

    query = User.outerjoin(Team).select().where(User.id == u.id)
    loader = User.load(team=Team.none_as_none(False))

    # TODO: this should fail in 1.0, please remove by then
    u = await query.gino.load(loader).first()
    assert u.team.id is None

    query = User.outerjoin(Team).select().where(User.id == u.id)
    loader = User.load(team=Team)

    u = await query.gino.load(loader).first()
    assert not hasattr(u, 'team')

    u = await User.load(team=Team).query.where(User.id == u.id).gino.first()
    assert not hasattr(u, 'team')

    query = User.outerjoin(Team).select().where(User.id == u.id)
    loader = User.load(team=Team.distinct(Team.id))

    u = await query.gino.load(loader).first()
    assert not hasattr(u, 'team')


async def test_tuple_loader_279(user):
    from gino.loader import TupleLoader
    query = db.select([User, Team])
    async with db.transaction():
        async for row in query.gino.load((User, Team)).iterate():
            assert len(row) == 2
        async for row in query.gino.load(TupleLoader((User, Team))).iterate():
            assert len(row) == 2


async def test_none_as_none_281(user):
    import gino

    if gino.__version__ < '0.9':
        query = Team.outerjoin(User).select()
        loader = Team, User.none_as_none()
        assert any(row[1] is None
                   for row in await query.gino.load(loader).all())

        loader = Team.distinct(Team.id).load(add_member=User.none_as_none())
        assert any(not team.members
                   for team in await query.gino.load(loader).all())

    if gino.__version__ >= '0.8.0':
        query = Team.outerjoin(User).select()
        loader = Team, User
        assert any(row[1] is None
                   for row in await query.gino.load(loader).all())

        loader = Team.distinct(Team.id).load(add_member=User)
        assert any(not team.members
                   for team in await query.gino.load(loader).all())



import pytest
import tornado.web
import tornado.httpclient
from tornado.ioloop import IOLoop
import tornado.options
import tornado.escape
from gino.ext.tornado import Gino, DBMixin, RequestHandlerMixin

from .models import DB_ARGS


# noinspection PyShadowingNames
@pytest.fixture
def app(ssl_ctx):
    # Define your database metadata
    # -----------------------------

    db = Gino()

    # Define tables as you would normally do
    # --------------------------------------

    class User(db.Model):
        __tablename__ = 'users'

        id = db.Column(db.Integer(), primary_key=True, autoincrement=True)
        nickname = db.Column(db.Unicode(), nullable=False)

    # Now just use your tables
    # ------------------------

    class AllUsers(tornado.web.RequestHandler):
        async def get(self):
            users = await User.query.gino.all()

            for user in users:
                self.write('<a href="{url}">{nickname}</a><br/>'.format(
                    url=self.application.reverse_url('user', user.id),
                    nickname=tornado.escape.xhtml_escape(user.nickname)))

    class GetUser(tornado.web.RequestHandler, RequestHandlerMixin):
        async def get(self, uid):
            async with self.db.acquire() as conn:
                user = await User.get_or_404(int(uid), bind=conn)
                self.write('Hi, {}!'.format(user.nickname))

    class AddUser(tornado.web.RequestHandler):
        async def post(self):
            user = await User.create(nickname=self.get_argument('name'))
            self.write('Hi, {}!'.format(user.nickname))

    class Application(tornado.web.Application, DBMixin):
        pass

    app = Application([
        tornado.web.URLSpec(r'/', AllUsers, name='index'),
        tornado.web.URLSpec(r'/user/(?P<uid>[0-9]+)', GetUser, name='user'),
        tornado.web.URLSpec(r'/user', AddUser, name='user_add'),
    ], debug=True)
    IOLoop.current().run_sync(
        lambda: app.init_db(db, host=DB_ARGS['host'], port=DB_ARGS['port'],
                            user=DB_ARGS['user'],
                            password=DB_ARGS['password'],
                            database=DB_ARGS['database'],
                            ssl=ssl_ctx))
    loop = tornado.ioloop.IOLoop.current().asyncio_loop
    loop.run_until_complete(db.gino.create_all())
    try:
        yield app
    finally:
        loop.run_until_complete(db.gino.drop_all())


@pytest.mark.gen_test
def test_hello_world(http_client, base_url):
    response = yield http_client.fetch(base_url)
    assert response.code == 200

    with pytest.raises(tornado.httpclient.HTTPClientError, match='404'):
        yield http_client.fetch(base_url + '/user/1')

    response = yield http_client.fetch(base_url + '/user', method='POST',
                                       body='name=fantix')
    assert response.code == 200
    assert b'fantix' in response.body

    response = yield http_client.fetch(base_url + '/user/1')
    assert response.code == 200
    assert b'fantix' in response.body

    response = yield http_client.fetch(base_url)
    assert response.code == 200
    assert b'fantix' in response.body



import pytest
import gino
from gino.declarative import InvertDict
from asyncpg.exceptions import (
    UniqueViolationError, ForeignKeyViolationError, CheckViolationError)

from .models import User, UserSetting

pytestmark = pytest.mark.asyncio
db = gino.Gino()


# noinspection PyUnusedLocal
async def test_column_not_deletable(bind):
    u = await User.create(nickname='test')
    with pytest.raises(AttributeError):
        del u.nickname


async def test_table_args():
    class Model(db.Model):
        __tablename__ = 'model1'

    assert Model.__table__.implicit_returning

    class Model(db.Model):
        __tablename__ = 'model2'

        __table_args__ = dict(implicit_returning=False)

    assert not Model.__table__.implicit_returning

    class Model(db.Model):
        __tablename__ = 'model3'

        __table_args__ = db.Column('new_col'), dict(implicit_returning=False)

    assert not Model.__table__.implicit_returning
    assert not hasattr(Model, 'new_col')
    assert not hasattr(Model.__table__.c, 'nonexist')
    assert hasattr(Model.__table__.c, 'new_col')

    class Model(db.Model):
        __tablename__ = 'model4'
        __table_args__ = db.Column('col1'), db.Column('col2')

        col3 = db.Column()

    assert not hasattr(Model, 'col1')
    assert not hasattr(Model, 'col2')
    assert hasattr(Model, 'col3')
    assert hasattr(Model.__table__.c, 'col1')
    assert hasattr(Model.__table__.c, 'col2')
    assert hasattr(Model.__table__.c, 'col3')


async def test_inline_constraints_and_indexes(bind, engine):
    u = await User.create(nickname='test')
    us1 = await UserSetting.create(user_id=u.id, setting='skin', value='blue')

    # PrimaryKeyConstraint
    with pytest.raises(UniqueViolationError):
        await UserSetting.create(id=us1.id, user_id=u.id, setting='key1',
                                 value='val1')

    # ForeignKeyConstraint
    with pytest.raises(ForeignKeyViolationError):
        await UserSetting.create(user_id=42, setting='key2', value='val2')

    # UniqueConstraint
    with pytest.raises(UniqueViolationError):
        await UserSetting.create(user_id=u.id, setting='skin',
                                 value='duplicate-setting')

    # CheckConstraint
    with pytest.raises(CheckViolationError):
        await UserSetting.create(user_id=u.id, setting='key3', value='val3',
                                 col1=42)

    # Index
    status, result = await engine.status(
        "SELECT * FROM pg_indexes WHERE indexname = 'col2_idx'")
    assert status == 'SELECT 1'


async def test_join_t112(engine):
    class Car(db.Model):
        __tablename__ = 'cars'

        id = db.Column(db.BigInteger(), primary_key=True)

    class Wheel(db.Model):
        __tablename__ = 'wheels'

        id = db.Column(db.BigInteger(), primary_key=True)
        car_id = db.Column(db.ForeignKey('cars.id'))

    sql = ('SELECT wheels.id, wheels.car_id, cars.id \nFROM wheels '
           'JOIN cars ON cars.id = wheels.car_id')

    assert engine.compile(Wheel.join(Car).select())[0] == sql


async def test_mixin():
    class Tracked:
        created = db.Column(db.DateTime(timezone=True))

        @db.declared_attr
        def unique_id(cls):
            return db.Column(db.Integer())

        @db.declared_attr
        def unique_constraint(cls):
            return db.UniqueConstraint('unique_id')

        @db.declared_attr
        def poly(cls):
            if cls.__name__ == 'Thing':
                return db.Column(db.Unicode())

        @db.declared_attr
        def __table_args__(cls):
            if cls.__name__ == 'Thing':
                return db.UniqueConstraint('poly'),

    class Audit(Tracked):
        pass

    class Thing(Audit, db.Model):
        __tablename__ = 'thing'

        id = db.Column(db.Integer, primary_key=True)

    class Another(Audit, db.Model):
        __tablename__ = 'another'

        id = db.Column(db.Integer, primary_key=True)

    assert isinstance(Thing.__table__.c.created, db.Column)
    assert isinstance(Another.__table__.c.created, db.Column)
    assert Thing.created is not Another.created
    assert Thing.created is Thing.__table__.c.created
    assert Another.created is Another.__table__.c.created

    assert Thing.unique_id is not Another.unique_id
    assert Thing.unique_id is Thing.__table__.c.unique_id
    c1, c2 = [list(filter(lambda c: list(c.columns)[0].name == 'unique_id',
                          m.__table__.constraints))[0]
              for m in [Thing, Another]]
    assert isinstance(c1, db.UniqueConstraint)
    assert isinstance(c2, db.UniqueConstraint)
    assert c1 is not c2

    assert isinstance(Thing.poly, db.Column)
    assert Another.poly is None
    for c in Thing.__table__.constraints:
        if list(c.columns)[0].name == 'poly':
            assert isinstance(c, db.UniqueConstraint)
            break
    else:
        assert False, 'Should not reach here'


# noinspection PyUnusedLocal
async def test_inherit_constraint():
    with pytest.raises(ValueError, match='already attached to another table'):
        class IllegalUserSetting(UserSetting):
            __table__ = None
            __tablename__ = 'bad_gino_user_settings'


async def test_abstract_model_error():
    class ConcreteModel(db.Model):
        __tablename__ = 'some_table'

        c = db.Column(db.Unicode())

    class AbstractModel(db.Model):
        pass

    with pytest.raises(TypeError, match='AbstractModel is abstract'):
        ConcreteModel.join(AbstractModel)

    with pytest.raises(TypeError, match='AbstractModel is abstract'):
        AbstractModel.join(ConcreteModel)

    with pytest.raises(TypeError, match='AbstractModel is abstract'):
        db.select(AbstractModel)

    with pytest.raises(TypeError, match='AbstractModel is abstract'):
        db.select([AbstractModel])

    with pytest.raises(TypeError, match='AbstractModel is abstract'):
        # noinspection PyStatementEffect
        AbstractModel.query

    with pytest.raises(TypeError, match='AbstractModel is abstract'):
        # noinspection PyStatementEffect
        AbstractModel.update

    am = AbstractModel()

    with pytest.raises(TypeError, match='AbstractModel is abstract'):
        await am.create()

    with pytest.raises(TypeError, match='AbstractModel is abstract'):
        await am.delete()

    req = am.update()

    with pytest.raises(TypeError, match='AbstractModel has no table'):
        await req.apply()

    with pytest.raises(TypeError, match='AbstractModel is abstract'):
        # noinspection PyStatementEffect
        AbstractModel.delete

    with pytest.raises(TypeError, match='AbstractModel is abstract'):
        AbstractModel.alias()

    with pytest.raises(TypeError, match='AbstractModel is abstract'):
        AbstractModel.alias()

    with pytest.raises(TypeError, match='AbstractModel is abstract'):
        await AbstractModel.get(1)


async def test_invert_dict():
    with pytest.raises(gino.GinoException,
                       match=r'Column name c1 already maps to \w+'):
        InvertDict({'col1': 'c1', 'col2': 'c1'})

    with pytest.raises(gino.GinoException,
                       match=r'Column name c1 already maps to \w+'):
        d = InvertDict()
        d['col1'] = 'c1'
        d['col2'] = 'c1'

    d = InvertDict()
    d['col1'] = 'c1'
    # it works for same key/value pair
    d['col1'] = 'c1'
    d['col2'] = 'c2'
    assert d.invert_get('c1') == 'col1'
    assert d.invert_get('c2') == 'col2'



import asyncio

from quart import Quart, request
from quart.exceptions import NotFound
from sqlalchemy.engine.url import URL

from ..api import Gino as _Gino, GinoExecutor as _Executor
from ..engine import GinoConnection as _Connection, GinoEngine as _Engine
from ..strategies import GinoStrategy


class QuartModelMixin:
    @classmethod
    async def get_or_404(cls, *args, **kwargs):
        # noinspection PyUnresolvedReferences
        rv = await cls.get(*args, **kwargs)
        if rv is None:
            raise NotFound()
        return rv


class GinoExecutor(_Executor):
    async def first_or_404(self, *args, **kwargs):
        rv = await self.first(*args, **kwargs)
        if rv is None:
            raise NotFound()
        return rv


class GinoConnection(_Connection):
    async def first_or_404(self, *args, **kwargs):
        rv = await self.first(*args, **kwargs)
        if rv is None:
            raise NotFound()
        return rv


class GinoEngine(_Engine):
    connection_cls = GinoConnection

    async def first_or_404(self, *args, **kwargs):
        rv = await self.first(*args, **kwargs)
        if rv is None:
            raise NotFound()
        return rv


class QuartStrategy(GinoStrategy):
    name = 'quart'
    engine_cls = GinoEngine


QuartStrategy()


# noinspection PyClassHasNoInit
class Gino(_Gino):
    """Support Quart web server.

    By :meth:`init_app` GINO registers a few hooks on Quart, so that GINO could
    use database configuration in Quart ``config`` to initialize the bound
    engine.

    A lazy connection context is enabled by default for every request. You can
    change this default behavior by setting ``DB_USE_CONNECTION_FOR_REQUEST``
    config value to ``False``. By default, a database connection is borrowed on
    the first query, shared in the same execution context, and returned to the
    pool on response. If you need to release the connection early in the middle
    to do some long-running tasks, you can simply do this::

        await request.connection.release(permanent=False)

    This doesn't apply to websocket, because websocket is usually a long
    connection, so it's not efficient to hold the connection.

    """
    model_base_classes = _Gino.model_base_classes + (QuartModelMixin,)
    query_executor = GinoExecutor

    def __init__(self, app=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Quart):
        if app.config.setdefault('DB_USE_CONNECTION_FOR_REQUEST', True):
            @app.before_request
            async def before_request():
                request.connection = await self.acquire(lazy=True)

            @app.after_request
            async def after_response(response):
                conn = getattr(request, 'connection', None)
                if conn is not None:
                    await conn.release()
                    del request.connection
                return response

        @app.before_first_request
        async def before_first_request():
            dsn = app.config.get('DB_DSN')
            if not dsn:
                dsn = URL(
                    drivername=app.config.setdefault('DB_DRIVER', 'asyncpg'),
                    host=app.config.setdefault('DB_HOST', 'localhost'),
                    port=app.config.setdefault('DB_PORT', 5432),
                    username=app.config.setdefault('DB_USER', 'postgres'),
                    password=app.config.setdefault('DB_PASSWORD', ''),
                    database=app.config.setdefault('DB_DATABASE', 'postgres'),
                )

            await self.set_bind(
                dsn,
                echo=app.config.setdefault('DB_ECHO', False),
                min_size=app.config.setdefault('DB_POOL_MIN_SIZE', 5),
                max_size=app.config.setdefault('DB_POOL_MAX_SIZE', 10),
                ssl=app.config.setdefault('DB_SSL'),
                loop=asyncio.get_event_loop(),
            )

    async def first_or_404(self, *args, **kwargs):
        rv = await self.first(*args, **kwargs)
        if rv is None:
            raise NotFound()
        return rv

    async def set_bind(self, bind, loop=None, **kwargs):
        kwargs.setdefault('strategy', 'quart')
        return await super().set_bind(bind, loop=loop, **kwargs)



from async_generator import yield_, async_generator
import pytest
import sanic
from sanic.response import text, json

import gino
from gino.ext.sanic import Gino

from .models import DB_ARGS, PG_URL


# noinspection PyShadowingNames
async def _app(config):
    app = sanic.Sanic()
    app.config.update(config)

    db = Gino(app)

    class User(db.Model):
        __tablename__ = 'gino_users'

        id = db.Column(db.BigInteger(), primary_key=True)
        nickname = db.Column(db.Unicode(), default='noname')

    @app.route('/')
    async def root(request):
        return text('Hello, world!')

    @app.route('/users/<uid:int>')
    async def get_user(request, uid):
        method = request.args.get('method')
        q = User.query.where(User.id == uid)
        if method == '1':
            return json((await q.gino.first_or_404()).to_dict())
        elif method == '2':
            return json(
                (await request['connection'].first_or_404(q)).to_dict())
        elif method == '3':
            return json(
                (await db.bind.first_or_404(q)).to_dict())
        elif method == '4':
            return json(
                (await db.first_or_404(q)).to_dict())
        else:
            return json((await User.get_or_404(uid)).to_dict())

    @app.route('/users', methods=['POST'])
    async def add_user(request):
        u = await User.create(nickname=request.form.get('name'))
        await u.query.gino.first_or_404()
        await db.first_or_404(u.query)
        await db.bind.first_or_404(u.query)
        await request['connection'].first_or_404(u.query)
        return json(u.to_dict())

    e = await gino.create_engine(PG_URL)
    try:
        try:
            await db.gino.create_all(e)
            await yield_(app)
        finally:
            await db.gino.drop_all(e)
    finally:
        await e.close()


@pytest.fixture
@async_generator
async def app():
    await _app({
        'DB_HOST': DB_ARGS['host'],
        'DB_PORT': DB_ARGS['port'],
        'DB_USER': DB_ARGS['user'],
        'DB_PASSWORD': DB_ARGS['password'],
        'DB_DATABASE': DB_ARGS['database'],
    })


@pytest.fixture
@async_generator
async def app_ssl(ssl_ctx):
    await _app({
        'DB_HOST': DB_ARGS['host'],
        'DB_PORT': DB_ARGS['port'],
        'DB_USER': DB_ARGS['user'],
        'DB_PASSWORD': DB_ARGS['password'],
        'DB_DATABASE': DB_ARGS['database'],
        'DB_SSL': ssl_ctx,
    })


@pytest.fixture
@async_generator
async def app_dsn():
    await _app({'DB_DSN': PG_URL})


def _test_index_returns_200(app):
    request, response = app.test_client.get('/')
    assert response.status == 200
    assert response.text == 'Hello, world!'


def test_index_returns_200(app):
    _test_index_returns_200(app)


def test_index_returns_200_dsn(app_dsn):
    _test_index_returns_200(app_dsn)


def _test(app):
    for method in '01234':
        request, response = app.test_client.get('/users/1?method=' + method)
        assert response.status == 404

    request, response = app.test_client.post('/users',
                                             data=dict(name='fantix'))
    assert response.status == 200
    assert response.json == dict(id=1, nickname='fantix')

    for method in '01234':
        request, response = app.test_client.get('/users/1?method=' + method)
        assert response.status == 200
        assert response.json == dict(id=1, nickname='fantix')


def test(app):
    _test(app)


def test_ssl(app_ssl):
    _test(app_ssl)


def test_dsn(app_dsn):
    _test(app_dsn)



# noinspection PyPackageRequirements
from aiohttp.web import HTTPNotFound, middleware
from sqlalchemy.engine.url import URL

from ..api import Gino as _Gino, GinoExecutor as _Executor
from ..engine import GinoConnection as _Connection, GinoEngine as _Engine
from ..strategies import GinoStrategy


class AiohttpModelMixin:
    @classmethod
    async def get_or_404(cls, *args, **kwargs):
        # noinspection PyUnresolvedReferences
        rv = await cls.get(*args, **kwargs)
        if rv is None:
            raise HTTPNotFound(reason='{} is not found'.format(cls.__name__))
        return rv


# noinspection PyClassHasNoInit
class GinoExecutor(_Executor):
    async def first_or_404(self, *args, **kwargs):
        rv = await self.first(*args, **kwargs)
        if rv is None:
            raise HTTPNotFound(reason='No such data')
        return rv


# noinspection PyClassHasNoInit
class GinoConnection(_Connection):
    async def first_or_404(self, *args, **kwargs):
        rv = await self.first(*args, **kwargs)
        if rv is None:
            raise HTTPNotFound(reason='No such data')
        return rv


# noinspection PyClassHasNoInit
class GinoEngine(_Engine):
    connection_cls = GinoConnection

    async def first_or_404(self, *args, **kwargs):
        rv = await self.first(*args, **kwargs)
        if rv is None:
            raise HTTPNotFound(reason='No such data')
        return rv


class AiohttpStrategy(GinoStrategy):
    name = 'aiohttp'
    engine_cls = GinoEngine


AiohttpStrategy()


@middleware
class Gino(_Gino):
    """Support aiohttp.web server.

    The common usage looks like this::

        from aiohttp import web
        from gino.ext.aiohttp import Gino

        db = Gino()
        app = web.Application(middlewares=[db])
        db.init_app(app)

    By :meth:`init_app` GINO subscribes to a few signals on aiohttp, so that
    GINO could use database configuration in aiohttp ``config.gino`` to
    initialize the bound engine. The config includes:

    * ``driver`` - the database driver, default is ``asyncpg``.
    * ``host`` - database server host, default is ``localhost``.
    * ``port`` - database server port, default is ``5432``.
    * ``user`` - database server user, default is ``postgres``.
    * ``password`` - database server password, default is empty.
    * ``database`` - database name, default is ``postgres``.
    * ``dsn`` - a SQLAlchemy database URL to create the engine, its existence
      will replace all previous connect arguments.
    * ``pool_min_size`` - the initial number of connections of the db pool.
    * ``pool_max_size`` - the maximum number of connections in the db pool.
    * ``echo`` - enable SQLAlchemy echo mode.
    * ``ssl`` - SSL context passed to ``asyncpg.connect``, default is ``None``.

    If the ``db`` is set as an aiohttp middleware, then a lazy connection is
    available at ``request['connection']``. By default, a database connection
    is borrowed on the first query, shared in the same execution context, and
    returned to the pool on response. If you need to release the connection
    early in the middle to do some long-running tasks, you can simply do this::

        await request['connection'].release(permanent=False)

    """
    model_base_classes = _Gino.model_base_classes + (AiohttpModelMixin,)
    query_executor = GinoExecutor

    def __call__(self, request, handler):
        return self._middleware(request, handler)

    async def _middleware(self, request, handler):
        async with self.acquire(lazy=True) as connection:
            request['connection'] = connection
            try:
                return await handler(request)
            finally:
                request.pop('connection', None)

    def init_app(self, app):
        app['db'] = self

        config = app['config'].get('gino', {})

        async def before_server_start(app_):
            if config.get('dsn'):
                dsn = config['dsn']
            else:
                dsn = URL(
                    drivername=config.setdefault('driver', 'asyncpg'),
                    host=config.setdefault('host', 'localhost'),
                    port=config.setdefault('port', 5432),
                    username=config.setdefault('user', 'postgres'),
                    password=config.setdefault('password', ''),
                    database=config.setdefault('database', 'postgres'),
                )

            await self.set_bind(
                dsn,
                echo=config.setdefault('echo', False),
                min_size=config.setdefault('pool_min_size', 5),
                max_size=config.setdefault('pool_max_size', 10),
                ssl=config.setdefault('ssl'),
                loop=app_.loop,
            )

        async def after_server_stop(app_):
            await self.pop_bind().close()

        app.on_startup.append(before_server_start)
        app.on_cleanup.append(after_server_stop)

    async def first_or_404(self, *args, **kwargs):
        rv = await self.first(*args, **kwargs)
        if rv is None:
            raise HTTPNotFound(reason='No such data')
        return rv

    async def set_bind(self, bind, loop=None, **kwargs):
        kwargs.setdefault('strategy', 'aiohttp')
        return await super().set_bind(bind, loop=loop, **kwargs)



import asyncio
import logging
from datetime import datetime

import asyncpg
from asyncpg.exceptions import InvalidCatalogNameError
from gino import UninitializedError
import pytest
from sqlalchemy.exc import ObjectNotExecutableError
import sqlalchemy as sa

from .models import db, User, PG_URL, qsize

pytestmark = pytest.mark.asyncio


async def test_basic(engine):
    init_size = qsize(engine)
    async with engine.acquire() as conn:
        assert isinstance(conn.raw_connection, asyncpg.Connection)
    assert init_size == qsize(engine)
    assert isinstance(await engine.scalar('select now()'), datetime)
    assert isinstance(await engine.scalar(sa.text('select now()')), datetime)
    assert isinstance((await engine.first('select now()'))[0], datetime)
    assert isinstance((await engine.all('select now()'))[0][0], datetime)
    status, result = await engine.status('select now()')
    assert status == 'SELECT 1'
    assert isinstance(result[0][0], datetime)
    with pytest.raises(ObjectNotExecutableError):
        await engine.all(object())


async def test_issue_79():
    import gino
    e = await gino.create_engine(PG_URL + '_non_exist', min_size=0)
    with pytest.raises(InvalidCatalogNameError):
        async with e.acquire():
            pass  # pragma: no cover
    # noinspection PyProtectedMember
    assert len(e._ctx.get([])) == 0


async def test_reuse(engine):
    init_size = qsize(engine)
    async with engine.acquire(reuse=True) as conn1:
        assert qsize(engine) == init_size - 1
        async with engine.acquire(reuse=True) as conn2:
            assert qsize(engine) == init_size - 1
            assert conn1.raw_connection is conn2.raw_connection
            assert await engine.scalar('select now()')
            assert qsize(engine) == init_size - 1
        assert qsize(engine) == init_size - 1
    assert qsize(engine) == init_size

    async with engine.acquire(reuse=False) as conn1:
        assert qsize(engine) == init_size - 1
        async with engine.acquire(reuse=True) as conn2:
            assert qsize(engine) == init_size - 1
            assert conn1.raw_connection is conn2.raw_connection
        assert qsize(engine) == init_size - 1
    assert qsize(engine) == init_size

    async with engine.acquire(reuse=True) as conn1:
        assert qsize(engine) == init_size - 1
        async with engine.acquire(reuse=False) as conn2:
            assert qsize(engine) == init_size - 2
            assert conn1.raw_connection is not conn2.raw_connection
        assert qsize(engine) == init_size - 1
    assert qsize(engine) == init_size

    async with engine.acquire(reuse=False) as conn1:
        assert qsize(engine) == init_size - 1
        async with engine.acquire(reuse=False) as conn2:
            assert qsize(engine) == init_size - 2
            assert conn1.raw_connection is not conn2.raw_connection
        assert qsize(engine) == init_size - 1
    assert qsize(engine) == init_size

    async with engine.acquire(reuse=False) as conn1:
        assert qsize(engine) == init_size - 1
        async with engine.acquire(reuse=True) as conn2:
            assert qsize(engine) == init_size - 1
            assert conn1.raw_connection is conn2.raw_connection
            async with engine.acquire(reuse=False) as conn3:
                assert qsize(engine) == init_size - 2
                assert conn1.raw_connection is not conn3.raw_connection
                async with engine.acquire(reuse=True) as conn4:
                    assert qsize(engine) == init_size - 2
                    assert conn3.raw_connection is conn4.raw_connection
                assert qsize(engine) == init_size - 2
            assert qsize(engine) == init_size - 1
        assert qsize(engine) == init_size - 1
    assert qsize(engine) == init_size


async def test_compile(engine):
    stmt, params = engine.compile(User.query.where(User.id == 3))
    assert params[0] == 3


async def test_logging(mocker):
    import gino
    mocker.patch('logging.Logger._log')
    sql = 'SELECT NOW() AS test_logging'

    e = await gino.create_engine(PG_URL, echo=False)
    await e.scalar(sql)
    await e.close()
    # noinspection PyProtectedMember,PyUnresolvedReferences
    logging.Logger._log.assert_not_called()

    e = await gino.create_engine(PG_URL, echo=True)
    await e.scalar(sql)
    await e.close()
    # noinspection PyProtectedMember,PyUnresolvedReferences
    logging.Logger._log.assert_any_call(logging.INFO, sql, ())


async def test_set_isolation_level():
    import gino
    with pytest.raises(sa.exc.ArgumentError):
        await gino.create_engine(PG_URL, isolation_level='non')
    e = await gino.create_engine(PG_URL,
                                 isolation_level='READ_UNCOMMITTED')
    async with e.acquire() as conn:
        assert await e.dialect.get_isolation_level(
            conn.raw_connection) == 'READ UNCOMMITTED'
    async with e.transaction(isolation='serializable') as tx:
        assert await e.dialect.get_isolation_level(
            tx.connection.raw_connection) == 'SERIALIZABLE'


async def test_too_many_engine_args():
    import gino
    with pytest.raises(TypeError):
        await gino.create_engine(PG_URL, non_exist=None)


# noinspection PyUnusedLocal
async def test_scalar_return_none(bind):
    assert await User.query.where(
        User.nickname == 'nonexist').gino.scalar() is None


async def test_asyncpg_0120(bind, mocker):
    # for asyncpg 0.12.0
    assert await bind.first('rollback') is None

    orig = getattr(asyncpg.Connection, '_do_execute')

    class Stmt:
        def __init__(self, stmt):
            self._stmt = stmt

        def _get_attributes(self):
            raise TypeError

    async def new(*args, **kwargs):
        result, stmt = await orig(*args, **kwargs)
        return result, Stmt(stmt)

    mocker.patch('asyncpg.Connection._do_execute', new=new)

    assert await bind.first('rollback') is None


async def test_asyncpg_0120_iterate(bind, mocker):
    async with bind.transaction():
        gen = await db.iterate('rollback')
        assert await gen.next() is None

    mocker.patch('asyncpg.prepared_stmt.'
                 'PreparedStatement.get_attributes').side_effect = TypeError

    async with bind.transaction():
        gen = await db.iterate('rollback')
        assert await gen.next() is None


async def test_async_metadata():
    import gino
    db_ = await gino.Gino(PG_URL)
    assert isinstance((await db_.scalar('select now()')), datetime)
    await db_.pop_bind().close()
    with pytest.raises(UninitializedError):
        db.bind.first()


# noinspection PyUnreachableCode
async def test_acquire_timeout():
    import gino
    e = await gino.create_engine(PG_URL, min_size=1, max_size=1)
    async with e.acquire():
        with pytest.raises(asyncio.TimeoutError):
            async with e.acquire(timeout=0.1):
                assert False, 'Should not reach here'

    loop = asyncio.get_event_loop()
    f1 = loop.create_future()

    async def first():
        async with e.acquire() as conn:
            f1.set_result(None)
            await asyncio.sleep(0.2)
            # noinspection PyProtectedMember
            return conn.raw_connection._con

    async def second():
        async with e.acquire(lazy=True) as conn:
            conn = conn.execution_options(timeout=0.1)
            with pytest.raises(asyncio.TimeoutError):
                assert await conn.scalar('select 1')

    async def third():
        async with e.acquire(reuse=True, timeout=0.4) as conn:
            # noinspection PyProtectedMember
            return conn.raw_connection._con

    t1 = loop.create_task(first())
    await f1
    loop.create_task(second())
    t3 = loop.create_task(third())
    assert await t1 is await t3


# noinspection PyProtectedMember
async def test_lazy(mocker):
    import gino
    engine = await gino.create_engine(PG_URL, min_size=1, max_size=1)
    init_size = qsize(engine)
    async with engine.acquire(lazy=True):
        assert qsize(engine) == init_size
        assert len(engine._ctx.get()) == 1
    assert engine._ctx.get() is None
    assert qsize(engine) == init_size
    async with engine.acquire(lazy=True):
        assert qsize(engine) == init_size
        assert len(engine._ctx.get()) == 1
        assert await engine.scalar('select 1')
        assert qsize(engine) == init_size - 1
        assert len(engine._ctx.get()) == 1
    assert engine._ctx.get() is None
    assert qsize(engine) == init_size

    loop = asyncio.get_event_loop()
    fut = loop.create_future()

    async def block():
        async with engine.acquire():
            fut.set_result(None)
            await asyncio.sleep(0.3)

    blocker = loop.create_task(block())
    await fut
    init_size_2 = qsize(engine)
    ctx = engine.acquire(lazy=True)
    conn = await ctx.__aenter__()
    t1 = loop.create_task(
        conn.execution_options(timeout=0.1).scalar('select 1'))
    t2 = loop.create_task(ctx.__aexit__(None, None, None))
    with pytest.raises(asyncio.TimeoutError):
        await t1
    assert not await t2
    assert qsize(engine) == init_size_2
    await blocker
    assert qsize(engine) == init_size

    fut = loop.create_future()
    blocker = loop.create_task(block())
    await fut
    init_size_2 = qsize(engine)

    async def acquire_failed(*args, **kwargs):
        await asyncio.sleep(0.1)
        raise ValueError()

    mocker.patch('asyncpg.pool.Pool.acquire', new=acquire_failed)
    ctx = engine.acquire(lazy=True)
    conn = await ctx.__aenter__()
    t1 = loop.create_task(conn.scalar('select 1'))
    t2 = loop.create_task(conn.release(permanent=False))
    with pytest.raises(ValueError):
        await t1
    assert not await t2
    assert qsize(engine) == init_size_2

    await conn.release(permanent=False)
    assert qsize(engine) == init_size_2

    await blocker
    assert qsize(engine) == init_size


async def test_release(engine):
    init_size = qsize(engine)
    async with engine.acquire() as conn:
        assert await conn.scalar('select 8') == 8
        await conn.release(permanent=False)
        assert await conn.scalar('select 8') == 8
        await conn.release(permanent=False)
    with pytest.raises(ValueError, match='released permanently'):
        await conn.scalar('select 8')
    with pytest.raises(ValueError, match='already released'):
        await conn.release()

    conn = await engine.acquire()
    assert await conn.scalar('select 8') == 8
    await conn.release(permanent=False)
    assert await conn.scalar('select 8') == 8
    await conn.release()
    with pytest.raises(ValueError, match='released permanently'):
        await conn.scalar('select 8')

    conn1 = await engine.acquire()
    conn2 = await engine.acquire(reuse=True)
    conn3 = await engine.acquire()
    conn4 = await engine.acquire(reuse=True)
    assert await conn1.scalar('select 8') == 8
    assert await conn2.scalar('select 8') == 8
    assert await conn3.scalar('select 8') == 8
    assert await conn4.scalar('select 8') == 8

    await conn1.release(permanent=False)
    assert await conn2.scalar('select 8') == 8

    await conn2.release(permanent=False)
    assert await conn2.scalar('select 8') == 8

    await conn1.release()
    with pytest.raises(ValueError, match='released permanently'):
        await conn2.scalar('select 8')
    assert await conn4.scalar('select 8') == 8

    await conn4.release()
    with pytest.raises(ValueError, match='released permanently'):
        await conn4.scalar('select 8')

    assert await conn3.scalar('select 8') == 8
    await conn3.release(permanent=False)
    assert await conn3.scalar('select 8') == 8
    assert init_size - 1 == qsize(engine)
    await conn3.release(permanent=False)
    assert init_size == qsize(engine)
    await conn3.release()
    assert init_size == qsize(engine)

    conn1 = await engine.acquire()
    conn2 = await engine.acquire()
    conn3 = await engine.acquire()
    assert engine.current_connection is conn3
    await conn2.release()
    assert engine.current_connection is conn3
    await conn1.release()
    assert engine.current_connection is conn3
    await conn3.release()
    assert engine.current_connection is None
    assert init_size == qsize(engine)


async def test_ssl():
    import ssl
    import gino

    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE

    e = await gino.create_engine(PG_URL, ssl=ctx)
    await e.close()


async def test_issue_313(bind):
    assert bind._ctx.get() is None

    async with db.acquire():
        pass

    assert bind._ctx.get() is None

    async def task():
        async with db.acquire(reuse=True):
            await db.scalar('SELECT now()')

    await asyncio.gather(*[task() for _ in range(5)])

    assert bind._ctx.get() is None

    async def task():
        async with db.transaction():
            await db.scalar('SELECT now()')

    await asyncio.gather(*[task() for _ in range(5)])

    assert bind._ctx.get() is None



import pytest
from datetime import datetime, timedelta

from .models import db, User, UserType

pytestmark = pytest.mark.asyncio


async def test_in_memory():
    u = User()
    assert u.age == 18
    u.age += 10
    assert u.age == 28
    assert u.balance == 0
    assert isinstance(u.balance, float)


# noinspection PyUnusedLocal
async def test_crud(bind):
    from gino.json_support import DATETIME_FORMAT
    now = datetime.utcnow()
    now_str = now.strftime(DATETIME_FORMAT)
    u = await User.create(nickname='fantix', birthday=now)
    u.age += 1
    assert await u.query.gino.model(None).first() == (
        1, 'fantix', {'age': 18, 'birthday': now_str}, UserType.USER, None)

    u = await User.get(u.id)
    assert u.nickname == 'fantix'
    assert u.birthday == now
    assert u.age == 18
    assert u.balance == 0
    assert isinstance(u.balance, float)
    assert await db.select([User.birthday]).where(
        User.id == u.id).gino.scalar() == now

    # In-memory update, not applying
    u.update(birthday=now - timedelta(days=3650))

    # Update two JSON fields, one using expression
    await u.update(age=User.age - 2, balance=100.85).apply()

    assert u.birthday == now - timedelta(days=3650)
    assert u.age == 16
    assert u.balance == 100
    assert isinstance(u.balance, float)
    assert await u.query.gino.model(None).first() == (
        1, 'fantix', dict(age=16, balance=100, birthday=now_str),
        UserType.USER, None)
    assert await db.select([User.realname]).where(
        User.id == u.id).gino.scalar() is None

    # Reload and test updating both JSON and regular property
    u = await User.get(u.id)
    await u.update(age=User.age - 2, balance=200.15, realname='daisy',
                   nickname='daisy.nick').apply()
    assert await u.query.gino.model(None).first() == (
        1, 'daisy.nick',
        dict(age=14, balance=200, realname='daisy', birthday=now_str),
        UserType.USER, None)
    assert u.to_dict() == dict(
        age=14,
        balance=200.0,
        birthday=now,
        id=1,
        nickname='daisy.nick',
        realname='daisy',
        type=UserType.USER,
        team_id=None,
    )

    # Deleting property doesn't affect database
    assert u.balance == 200
    u.balance = 300
    assert u.balance == 300
    del u.balance
    assert u.balance == 0
    assert await db.select([User.balance]).where(
        User.id == u.id).gino.scalar() == 200
    await u.update(age=22).apply()
    assert u.balance == 0
    assert await db.select([User.balance]).where(
        User.id == u.id).gino.scalar() == 200
    await u.update(balance=None).apply()
    assert u.balance == 0
    assert await db.select([User.balance]).where(
        User.id == u.id).gino.scalar() is None


# noinspection PyUnusedLocal
async def test_non_jsonb(bind):
    from gino.dialects.asyncpg import JSON

    class News(db.Model):
        __tablename__ = 'news'

        id = db.Column(db.Integer(), primary_key=True)
        profile = db.Column(JSON(), nullable=False, server_default='{}')
        visits = db.IntegerProperty(default=0)

    await News.gino.create()
    try:
        news = await News.create()
        assert news.visits == 0
        with pytest.raises(TypeError, match='JSON is not supported'):
            await news.update(visits=News.visits + 10).apply()
        assert news.visits == 0
        with pytest.raises(TypeError, match='JSON is not supported'):
            await news.update(visits=10).apply()
        assert news.visits == 10
        assert await news.select('visits').gino.scalar() == 0
        await news.update(profile=dict(visits=20)).apply()
        assert news.visits == 10
        assert await news.select('visits').gino.scalar() == 20
    finally:
        await News.gino.drop()


# noinspection PyUnusedLocal
async def test_reload(bind):
    u = await User.create()
    await u.update(realname=db.cast('888', db.Unicode)).apply()
    assert u.realname == '888'
    await u.update(profile=None).apply()
    assert u.realname == '888'
    User.__dict__['realname'].reload(u)
    assert u.realname is None


# noinspection PyUnusedLocal
async def test_properties(bind):
    from gino.dialects.asyncpg import JSONB

    class PropsTest(db.Model):
        __tablename__ = 'props_test'
        profile = db.Column(JSONB(), nullable=False, server_default='{}')

        raw = db.JSONProperty()
        bool = db.BooleanProperty()
        obj = db.ObjectProperty()
        arr = db.ArrayProperty()

    await PropsTest.gino.create()
    try:
        t = await PropsTest.create(
            raw=dict(a=[1, 2]),
            bool=True,
            obj=dict(x=1, y=2),
            arr=[3, 4, 5, 6],
        )
        assert t.obj['x'] == 1
        assert t.arr[-1] == 6
        assert await db.select([
            PropsTest.profile,
            PropsTest.raw,
            PropsTest.bool,
        ]).gino.first() == (
                   {'arr': [3, 4, 5, 6], 'obj': {'x': 1, 'y': 2},
                    'raw': {'a': [1, 2]}, 'bool': True},
                   dict(a=[1, 2]),
                   True)
        t.obj = dict(x=10, y=20)
        assert t.obj['x'] == 10
        t.arr = [4, 5, 6, 7]
        assert t.arr[-1] == 7
    finally:
        await PropsTest.gino.drop()


async def test_no_profile():
    with pytest.raises(AttributeError, match=r'JSON\[B\] column'):
        # noinspection PyUnusedLocal
        class Test(db.Model):
            __tablename__ = 'tests_no_profile'

            age = db.IntegerProperty(default=18)


async def test_t291_t402(bind):
    from gino.dialects.asyncpg import JSON, JSONB

    class CustomJSONB(db.TypeDecorator):
        impl = JSONB

        def process_result_value(self, *_):
            return 123

    class PropsTest(db.Model):
        __tablename__ = 'props_test_291'
        profile = db.Column(JSONB(), nullable=False, server_default='{}')
        profile1 = db.Column(JSON(), nullable=False, server_default='{}')
        profile2 = db.Column(CustomJSONB(),
                             nullable=False, server_default='{}')

        bool = db.BooleanProperty()
        bool1 = db.BooleanProperty(prop_name='profile1')

    await PropsTest.gino.create()
    try:
        t = await PropsTest.create(
            bool=True,
            bool1=True,
        )
        profile = await bind.scalar('SELECT profile FROM props_test_291')
        assert isinstance(profile, dict)
        profile1 = await bind.scalar('SELECT profile1 FROM props_test_291')
        assert isinstance(profile1, dict)
        profile2 = await bind.scalar('SELECT profile2 FROM props_test_291')
        assert isinstance(profile2, dict)
        custom_profile2 = await bind.scalar(PropsTest.select('profile2'))
        assert isinstance(custom_profile2, int)
        assert custom_profile2 == 123
    finally:
        await PropsTest.gino.drop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst', encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst', encoding='utf-8') as history_file:
    history = history_file.read()


def req_file(filename):
    with open(filename) as f:
        content = f.readlines()
    return [x.strip() for x in content if x.strip()]


setup_requirements = [
    'pytest-runner',
]

test_requirements = [
    'pytest',
    'pytest-asyncio',
    'pytest-mock',
    'psycopg2-binary',
]

setup(
    name='gino',
    version='0.8.1',
    description="GINO Is Not ORM - "
                "a Python asyncio ORM on SQLAlchemy core.",
    long_description=readme + '\n\n' + history,
    author="Fantix King",
    author_email='fantix.king@gmail.com',
    url='https://github.com/fantix/gino',
    packages=[p for p in find_packages() if p != 'tests'],
    include_package_data=True,
    install_requires=req_file('requirements.txt'),
    python_requires='>=3.5',
    entry_points="""
    [sqlalchemy.dialects]
    postgresql.asyncpg = gino.dialects.asyncpg:AsyncpgDialect
    asyncpg = gino.dialects.asyncpg:AsyncpgDialect
    """,
    license="BSD license",
    zip_safe=False,
    keywords='orm asyncio sqlalchemy asyncpg python3 sanic aiohttp tornado',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    project_urls={
        'Documentation': 'https://python-gino.readthedocs.io/',
        'Funding': 'https://www.patreon.com/fantixking',
        'Source': 'https://github.com/fantix/gino',
        'Tracker': 'https://github.com/fantix/gino/issues',
    },
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



import pytest

from .models import db, User

pytestmark = pytest.mark.asyncio


# noinspection PyUnusedLocal
async def test_status(bind):
    statement, params = db.compile(User.insert(),
                                   [dict(name='1'), dict(name='2')])
    assert statement == ('INSERT INTO gino_users (name, type) '
                         'VALUES ($1, $2)')
    assert params == (('1', 'USER'), ('2', 'USER'))
    result = await User.insert().gino.status(dict(name='1'), dict(name='2'))
    assert result is None
    assert len(await User.query.gino.all()) == 2


# noinspection PyUnusedLocal
async def test_all(bind):
    result = await User.insert().returning(User.nickname).gino.all(
        dict(name='1'), dict(name='2'))
    assert result is None
    rows = await User.query.gino.all()
    assert len(rows) == 2
    assert set(u.nickname for u in rows) == {'1', '2'}

    result = await User.insert().gino.all(
        dict(name='3'), dict(name='4'))
    assert result is None
    rows = await User.query.gino.all()
    assert len(rows) == 4
    assert set(u.nickname for u in rows) == {'1', '2', '3', '4'}


# noinspection PyUnusedLocal
async def test_first(bind):
    result = await User.insert().returning(User.nickname).gino.first(
        dict(name='1'), dict(name='2'))
    assert result is None
    rows = await User.query.gino.all()
    assert len(await User.query.gino.all()) == 2
    assert set(u.nickname for u in rows) == {'1', '2'}

    result = await User.insert().gino.first(
        dict(name='3'), dict(name='4'))
    assert result is None
    rows = await User.query.gino.all()
    assert len(rows) == 4
    assert set(u.nickname for u in rows) == {'1', '2', '3', '4'}


# noinspection PyUnusedLocal
async def test_scalar(bind):
    result = await User.insert().returning(User.nickname).gino.scalar(
        dict(name='1'), dict(name='2'))
    assert result is None
    rows = await User.query.gino.all()
    assert len(await User.query.gino.all()) == 2
    assert set(u.nickname for u in rows) == {'1', '2'}

    result = await User.insert().gino.scalar(
        dict(name='3'), dict(name='4'))
    assert result is None
    rows = await User.query.gino.all()
    assert len(rows) == 4
    assert set(u.nickname for u in rows) == {'1', '2', '3', '4'}



"""
GINO provides a convenient plugin for integrating with Tornado_ webserver.

.. _Tornado: http://www.tornadoweb.org


Provide tornado-specific methods on models
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

GINO can provide a web framework-aware ``.get_or_404()`` method which work
similar to ``.get()`` but raises an appropriate error whenever requested object
not found. In case of tornado, an appropriate error is
``tornado.web.HTTPError(404)``.

To have it working, simply use :py:class:`gino.ext.tornado.Gino` as your
database metadata.


Integrate GINO with application
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

GINO provides two ways to initialize the db instances for tornado:

1) Initialize by ``db.init_app(app)``, and get the ``Gino`` instance in the
``RequestHandler`` subclasses by ``self.application.db``.

2) For subclassed tornado Application, use :py:class:`gino.ext.tornado.DBMixin`
and run :py:meth:`init_db`.


Request Handler Mixin
^^^^^^^^^^^^^^^^^^^^^

A mixin to provide a convenience property access to db using ``self.db``
instead of ``self.application.db``.


An example application
^^^^^^^^^^^^^^^^^^^^^^

A hello world application that uses tornado and GINO may look like this:

.. code-block:: python

    import ssl

    import tornado.web
    import tornado.ioloop
    import tornado.options
    import tornado.escape

    from gino.ext.tornado import Gino, RequestHandlerMixin

    # Define your database metadata
    # -----------------------------

    db = Gino()


    # Define tables as you would normally do
    # --------------------------------------

    class User(db.Model):
        __tablename__ = 'users'

        id = db.Column(db.Integer(), primary_key=True, autoincrement=True)
        nickname = db.Column(db.Unicode(), nullable=False)


    # Now just use your tables
    # ------------------------

    class AllUsers(tornado.web.RequestHandler, RequestHandlerMixin):
        async def get(self):
            users = await User.query.gino.all()

            for user in users:
                url = self.application.reverse_url('user', user.id)
                nickname = tornado.escape.xhtml_escape(user.nickname)
                self.write(f'<a href="{url}">{nickname}</a><br/>')


    class GetUser(tornado.web.RequestHandler, RequestHandlerMixin):
        async def get(self, uid):
            async with self.db.acquire() as conn:
                async with conn.transaction():
                    user: User = await User.get_or_404(int(uid))
                    self.write(f'Hi, {user.nickname}!')


    if __name__ == '__main__':
        app = tornado.web.Application([
            tornado.web.URLSpec(r'/', AllUsers, name='index'),
            tornado.web.URLSpec(r'/users/(?P<uid>[0-9]+)', GetUser,
                                name='user')
        ], debug=True)
        tornado.ioloop.IOLoop.current().run_sync(
            lambda: db.init_app(app, ssl=True))
        app.listen(8888)
        tornado.ioloop.IOLoop.current().start()

API reference
^^^^^^^^^^^^^

"""

import asyncio
import typing

import tornado
import tornado.ioloop
import tornado.iostream
import tornado.log
import tornado.platform.asyncio
import tornado.web

from sqlalchemy.engine.url import URL

from ..api import Gino as _Gino, GinoExecutor as _Executor
from ..engine import GinoConnection as _Connection, GinoEngine as _Engine
from ..strategies import GinoStrategy

if tornado.version_info[0] < 5:
    raise Exception('Only Tornado 5 or later is supported')


class TornadoModelMixin:
    @classmethod
    async def get_or_404(cls, *args, **kwargs):
        # noinspection PyUnresolvedReferences
        rv = await cls.get(*args, **kwargs)
        if rv is None:
            raise tornado.web.HTTPError(404)
        return rv


class GinoExecutor(_Executor):
    async def first_or_404(self, *args, **kwargs):
        rv = await self.first(*args, **kwargs)
        if rv is None:
            raise tornado.web.HTTPError(404)
        return rv


class GinoConnection(_Connection):
    async def first_or_404(self, *args, **kwargs):
        rv = await self.first(*args, **kwargs)
        if rv is None:
            raise tornado.web.HTTPError(404)
        return rv


class GinoEngine(_Engine):
    connection_cls = GinoConnection

    async def first_or_404(self, *args, **kwargs):
        rv = await self.first(*args, **kwargs)
        if rv is None:
            raise tornado.web.HTTPError(404)
        return rv


class TornadoStrategy(GinoStrategy):
    name = 'tornado'
    engine_cls = GinoEngine


TornadoStrategy()


class Gino(_Gino):
    """
    Base class for GINO database.

    Using this class as a metadata for your database adds an additional
    ``get_or_404()`` method to all of your table classes.

    """

    model_base_classes = _Gino.model_base_classes + (TornadoModelMixin,)
    query_executor = GinoExecutor

    if typing.TYPE_CHECKING:
        # Typehints to enable autocompletion on all Gino.Model-derived classes

        from ..crud import CRUDModel as __CRUDModel
        from ..declarative import ModelType as __ModelType

        class Model(__CRUDModel, TornadoModelMixin, metaclass=__ModelType):
            ...

    async def first_or_404(self, *args, **kwargs):
        rv = await self.first(*args, **kwargs)
        if rv is None:
            raise tornado.web.HTTPError(404)
        return rv

    async def set_bind(self, bind, loop=None, **kwargs):
        kwargs.setdefault('strategy', 'tornado')
        return await super().set_bind(bind, loop=loop, **kwargs)

    async def init_app(self, app, *, loop=None, dsn='', driver='asyncpg',
                       host='localhost', port=5432,
                       user='postgres', password='', database='postgres',
                       echo=False, pool_min_size=5, pool_max_size=10,
                       ssl=None):
        """
        Initialize database

        :param app: tornado.web.Application
        :param loop: User-defined event loop. If not defined, tornado default
        loop will be used.
        :param driver: the database driver, default is ``asyncpg``.
        :param host: database server host, default is ``localhost``.
        :param port: database server port, default is ``5432``.
        :param user: database server user, default is ``postgres``.
        :param password: database server password, default is empty.
        :param database: database name, default is ``postgres``.
        :param dsn: a SQLAlchemy database URL to create the engine, its
        existence will replace all previous connect arguments.
        :param echo: enable SQLAlchemy echo mode.
        :param pool_min_size: the initial number of connections of the db
        pool, default is ``5``.
        :param pool_max_size: the maximum number of connections in the db
        pool, default is ``10``.
        :param ssl: SSL context passed to ``asyncpg.connect``, default is
        ``None``. This can be ``True`` or an instance of ``ssl.SSLContext``.
        """
        if loop is None:
            loop = tornado.ioloop.IOLoop.current()
        if isinstance(loop, tornado.platform.asyncio.BaseAsyncIOLoop):
            asyncio_loop = loop.asyncio_loop
        elif isinstance(loop, asyncio.BaseEventLoop):
            asyncio_loop = loop
        else:
            raise RuntimeError('AsyncIOLoop is required to run GINO')

        if not dsn:
            dsn = URL(
                drivername=driver, host=host, port=port, username=user,
                password=password, database=database,
            )

        await self.set_bind(
            dsn, echo=echo, min_size=pool_min_size, max_size=pool_max_size,
            ssl=ssl, loop=asyncio_loop,
        )

        app.db = self


class DBMixin:
    """
    A mixin for tornado.web.Application to initialize and have convenience
    methods for database accesses.
    """

    db = None  # type: Gino

    async def init_db(self: [tornado.web.Application, 'DBMixin'],
                      db: Gino, **kwargs):
        await db.init_app(self, **kwargs)


class RequestHandlerMixin:
    """
    A mixin to provide convenience methods to access GINO object
    """

    @property
    def db(self: tornado.web.RequestHandler) -> Gino:
        return self.application.db



from enum import Enum

import pytest

import gino
from gino.dialects.asyncpg import AsyncEnum

pytestmark = pytest.mark.asyncio
db = gino.Gino()


class MyEnum(Enum):
    ONE = 'one'
    TWO = 'two'


class Blog(db.Model):
    __tablename__ = 's_blog'

    id = db.Column(db.BigInteger(), primary_key=True)
    title = db.Column(db.Unicode(), index=True, comment='Title Comment')
    visits = db.Column(db.BigInteger(), default=0)
    comment_id = db.Column(db.ForeignKey('s_comment.id'))
    number = db.Column(db.Enum(MyEnum), nullable=False, default=MyEnum.TWO)
    number2 = db.Column(AsyncEnum(MyEnum), nullable=False, default=MyEnum.TWO)


class Comment(db.Model):
    __tablename__ = 's_comment'

    id = db.Column(db.BigInteger(), primary_key=True)
    blog_id = db.Column(db.ForeignKey('s_blog.id', name='blog_id_fk'))


blog_seq = db.Sequence('blog_seq', metadata=db, schema='schema_test')


async def test(engine, define=True):
    try:
        async with engine.acquire() as conn:
            assert not await engine.dialect.has_schema(conn, 'schema_test')
            assert not await engine.dialect.has_table(conn, 'non_exist')
            assert not await engine.dialect.has_sequence(conn, 'non_exist')
            assert not await engine.dialect.has_type(conn, 'non_exist')
            assert not await engine.dialect.has_type(conn, 'non_exist',
                                                     schema='schema_test')
        await engine.status('create schema schema_test')
        Blog.__table__.schema = 'schema_test'
        Blog.__table__.comment = 'Blog Comment'
        Comment.__table__.schema = 'schema_test'
        db.bind = engine
        await db.gino.create_all()
        await Blog.number.type.create_async(engine, checkfirst=True)
        await Blog.number2.type.create_async(engine, checkfirst=True)
        await db.gino.create_all(tables=[Blog.__table__], checkfirst=True)
        await blog_seq.gino.create(checkfirst=True)
        await Blog.__table__.gino.create(checkfirst=True)
        await db.gino.drop_all()
        await db.gino.drop_all(tables=[Blog.__table__], checkfirst=True)
        await Blog.__table__.gino.drop(checkfirst=True)
        await blog_seq.gino.drop(checkfirst=True)

        if define:
            class Comment2(db.Model):
                __tablename__ = 's_comment_2'

                id = db.Column(db.BigInteger(), primary_key=True)
                blog_id = db.Column(db.ForeignKey('s_blog.id'))

            Comment2.__table__.schema = 'schema_test'
        await db.gino.create_all()
        await db.gino.drop_all()
    finally:
        await engine.status('drop schema schema_test cascade')


async def test_no_alter(engine, mocker):
    engine.dialect.supports_alter = False
    warn = mocker.patch('warnings.warn')
    await test(engine, define=False)
    assert warn.called



import gino
import pytest
from aiohttp import web
from async_generator import yield_, async_generator
from gino.ext.aiohttp import Gino

from .models import DB_ARGS, PG_URL

pytestmark = pytest.mark.asyncio


# noinspection PyShadowingNames
async def _test_client(config):
    from aiohttp.test_utils import TestServer, TestClient

    db = Gino()
    app = web.Application(middlewares=[db])
    app['config'] = dict(gino=config)
    db.init_app(app)

    class User(db.Model):
        __tablename__ = 'gino_users'

        id = db.Column(db.BigInteger(), primary_key=True)
        nickname = db.Column('name', db.Unicode(), default='noname')

    routes = web.RouteTableDef()

    @routes.get('/')
    async def root(request):
        return web.Response(text='Hello, world!')

    @routes.get('/users/{uid}')
    async def get_user(request):
        uid = int(request.match_info['uid'])
        method = request.query.get('method')
        q = User.query.where(User.id == uid)
        if method == '1':
            return web.json_response((await q.gino.first_or_404()).to_dict())
        elif method == '2':
            return web.json_response(
                (await request['connection'].first_or_404(q)).to_dict())
        elif method == '3':
            return web.json_response(
                (await db.bind.first_or_404(q)).to_dict())
        elif method == '4':
            return web.json_response(
                (await request.app['db'].first_or_404(q)).to_dict())
        else:
            return web.json_response((await User.get_or_404(uid)).to_dict())

    @routes.post('/users')
    async def add_user(request):
        form = await request.post()
        u = await User.create(nickname=form.get('name'))
        await u.query.gino.first_or_404()
        await db.first_or_404(u.query)
        await db.bind.first_or_404(u.query)
        await request['connection'].first_or_404(u.query)
        return web.json_response(u.to_dict())

    app.router.add_routes(routes)

    e = await gino.create_engine(PG_URL)
    try:
        try:
            await db.gino.create_all(e)
            async with TestClient(TestServer(app)) as rv:
                await yield_(rv)
        finally:
            await db.gino.drop_all(e)
    finally:
        await e.close()


@pytest.fixture
@async_generator
async def test_client():
    await _test_client(DB_ARGS.copy())


@pytest.fixture
@async_generator
async def test_client_dsn():
    await _test_client(dict(dsn=PG_URL))


@pytest.fixture
@async_generator
async def test_client_ssl(ssl_ctx):
    await _test_client(dict(dsn=PG_URL, ssl=ssl_ctx))


async def _test_index_returns_200(test_client):
    response = await test_client.get('/')
    assert response.status == 200
    assert await response.text() == 'Hello, world!'


async def test_index_returns_200(test_client):
    await _test_index_returns_200(test_client)


async def test_index_returns_200_dsn(test_client_dsn):
    await _test_index_returns_200(test_client_dsn)


async def _test(test_client):
    response = await test_client.get('/users/1')
    assert response.status == 404

    for method in '1234':
        response = await test_client.get('/users/1?method=' + method)
        assert response.status == 404

    response = await test_client.post('/users', data=dict(name='fantix'))
    assert response.status == 200
    assert await response.json() == dict(id=1, nickname='fantix')

    response = await test_client.get('/users/1')
    assert response.status == 200
    assert await response.json() == dict(id=1, nickname='fantix')


async def test(test_client):
    await _test(test_client)


async def test_dsn(test_client_dsn):
    await _test(test_client_dsn)


async def test_ssl(test_client_ssl):
    await _test(test_client_ssl)



import sys

import pytest

# Quart only supports Python 3.6 or later
if sys.version_info < (3, 6):
    raise pytest.skip(allow_module_level=True)

import json

from async_generator import yield_, async_generator
from quart import Quart, jsonify, request, websocket
from quart.exceptions import NotFound

import gino
from gino.ext.quart import Gino

from .models import DB_ARGS, PG_URL

pytestmark = pytest.mark.asyncio


# noinspection PyShadowingNames
async def _app(config):
    app = Quart(__name__)
    app.config.update(config)

    db = Gino(app)

    class User(db.Model):
        __tablename__ = 'gino_users'

        id = db.Column(db.BigInteger(), primary_key=True)
        nickname = db.Column(db.Unicode(), default='noname')

    @app.route('/')
    async def root():
        return 'Hello, world!'

    async def _get_user(ctx, uid: int, method: str) -> dict:
        q = User.query.where(User.id == uid)
        if method == '1':
            return (await q.gino.first_or_404()).to_dict()
        elif method == '2' and ctx:
            return (await ctx.connection.first_or_404(q)).to_dict()
        elif method == '3':
            return (await db.bind.first_or_404(q)).to_dict()
        elif method == '4':
            return (await db.first_or_404(q)).to_dict()
        else:
            return (await User.get_or_404(uid)).to_dict()

    @app.route('/users/<int:uid>')
    async def get_user(uid):
        method = request.args.get('method')
        return jsonify(await _get_user(request, uid, method))

    async def _add_user(ctx, nickname: str) -> dict:
        u = await User.create(nickname=nickname)
        await u.query.gino.first_or_404()
        await db.first_or_404(u.query)
        await db.bind.first_or_404(u.query)
        if ctx:
            await ctx.connection.first_or_404(u.query)
        return u.to_dict()

    @app.route('/users', methods=['POST'])
    async def add_user():
        return jsonify(await _add_user(
            request, (await request.form).get('name')))

    @app.websocket('/ws')
    async def ws():
        while True:
            data = json.loads(await websocket.receive())
            action = data.get('action')
            if action == 'add':
                new_user = await _add_user(None, data.get('name'))
                await websocket.send(json.dumps(new_user))
            elif action == 'get':
                try:
                    user = await _get_user(
                        None, int(data.get('id')), data.get('method'))
                except NotFound:
                    await websocket.send(json.dumps({'error': 'not found'}))
                else:
                    await websocket.send(json.dumps(user))
            else:
                await websocket.send(json.dumps({'error': 'Invalid JSON'}))

    e = await gino.create_engine(PG_URL)
    try:
        try:
            await db.gino.create_all(e)
            await yield_(app)
        finally:
            await db.gino.drop_all(e)
    finally:
        await e.close()


@pytest.fixture
@async_generator
async def app():
    await _app({
        'DB_HOST': DB_ARGS['host'],
        'DB_PORT': DB_ARGS['port'],
        'DB_USER': DB_ARGS['user'],
        'DB_PASSWORD': DB_ARGS['password'],
        'DB_DATABASE': DB_ARGS['database'],
    })


@pytest.fixture
@async_generator
async def app_ssl(ssl_ctx):
    await _app({
        'DB_HOST': DB_ARGS['host'],
        'DB_PORT': DB_ARGS['port'],
        'DB_USER': DB_ARGS['user'],
        'DB_PASSWORD': DB_ARGS['password'],
        'DB_DATABASE': DB_ARGS['database'],
        'DB_SSL': ssl_ctx,
    })


@pytest.fixture
@async_generator
async def app_dsn():
    await _app({'DB_DSN': PG_URL})


async def _test_index_returns_200(app):
    response = await app.test_client().get('/')
    assert response.status_code == 200
    assert await response.get_data(raw=False) == 'Hello, world!'


async def test_index_returns_200(app):
    await _test_index_returns_200(app)


async def test_index_returns_200_dsn(app_dsn):
    await _test_index_returns_200(app_dsn)


async def _test(app):
    test_client = app.test_client()
    for method in '01234':
        response = await test_client.get(
            '/users/1', query_string={'method': method})
        assert response.status_code == 404

    response = await test_client.post('/users', form=dict(name='fantix'))
    assert response.status_code == 200
    assert await response.get_json() == dict(id=1, nickname='fantix')

    for method in '01234':
        response = await test_client.get(
            '/users/1', query_string={'method': method})
        assert response.status_code == 200
        assert await response.get_json() == dict(id=1, nickname='fantix')


async def test(app):
    await _test(app)


async def test_ssl(app_ssl):
    await _test(app_ssl)


async def test_dsn(app_dsn):
    await _test(app_dsn)


async def _websocket_request(ws, req: dict) -> dict:
    await ws.send(json.dumps(req).encode('utf-8'))
    return json.loads(await ws.receive())


async def _test_websocket(app):
    test_client = app.test_client()
    with test_client.websocket('/ws') as ws:
        for method in '01234':
            response = await _websocket_request(
                ws, {'action': 'get', 'id': 1, 'method': method})
            assert response.get('error') == 'not found'

        response = await _websocket_request(
            ws, {'action': 'add', 'name': 'fantix'})
        assert response == dict(id=1, nickname='fantix')

        for method in '01234':
            response = await _websocket_request(
                ws, {'action': 'get', 'id': 1, 'method': method})
            assert response == dict(id=1, nickname='fantix')


async def test_ws(app):
    await _test_websocket(app)


async def test_ws_dsn(app_dsn):
    await _test_websocket(app_dsn)



import random

import pytest

from .models import db, User, UserType, Friendship, Relation, PG_URL

pytestmark = pytest.mark.asyncio


async def test_create(engine):
    nickname = 'test_create_{}'.format(random.random())
    u = await User.create(bind=engine, timeout=10,
                          nickname=nickname, age=42, type=UserType.USER)
    assert u.id is not None
    assert u.nickname == nickname
    assert u.type == UserType.USER
    assert u.age == 42

    u2 = await User.get(u.id, bind=engine, timeout=10)
    assert u2.id == u.id
    assert u2.nickname == nickname
    assert u2.type == UserType.USER
    assert u2.age == 42
    assert u2 is not u

    return u


async def test_create_from_instance(engine):
    nickname = 'test_create_from_instance_{}'.format(random.random())
    u = User(nickname='will-be-replaced', type=UserType.USER, age=42)
    u.nickname = nickname
    u.age = 21
    await u.create(bind=engine, timeout=10)
    assert u.id is not None
    assert u.nickname == nickname
    assert u.type == UserType.USER
    assert u.age == 21

    u2 = await User.get(u.id, bind=engine, timeout=10)
    assert u2.id == u.id
    assert u2.nickname == nickname
    assert u2.type == UserType.USER
    assert u2.age == 21
    assert u2 is not u

    return u


async def test_get(engine):
    u1 = await test_create(engine)
    u2 = await User.get(u1.id, bind=engine, timeout=10)
    assert u1.id == u2.id
    assert u1.nickname == u2.nickname
    assert u1 is not u2

    u3 = await engine.first(u1.query)
    assert u1.id == u3.id
    assert u1.nickname == u3.nickname
    assert u1 is not u3

    u4 = await test_create_from_instance(engine)
    u5 = await engine.first(u4.query)
    assert u4.id == u5.id
    assert u4.nickname == u5.nickname
    assert u4 is not u5


async def test_textual_sql(engine):
    u1 = await test_create(engine)
    u2 = await engine.first(db.text(
        'SELECT * FROM gino_users WHERE id = :uid'
    ).bindparams(uid=u1.id).columns(*User).execution_options(model=User))
    assert isinstance(u2, User)
    assert u1.id == u2.id
    assert u1.nickname == u2.nickname
    assert u1.type is u2.type
    assert u1 is not u2

    u2 = await engine.first(db.text(
        'SELECT * FROM gino_users WHERE id = :uid AND type = :utype'
    ).bindparams(
        db.bindparam('utype', type_=db.Enum(UserType))
    ).bindparams(
        uid=u1.id,
        utype=UserType.USER,
    ).columns(*User).execution_options(model=User))
    assert isinstance(u2, User)
    assert u1.id == u2.id
    assert u1.nickname == u2.nickname
    assert u1.type is u2.type
    assert u1 is not u2


async def test_select(engine):
    u = await test_create(engine)
    name = await engine.scalar(User.select('nickname').where(User.id == u.id))
    assert u.nickname == name

    name = await engine.scalar(u.select('nickname'))
    assert u.nickname == name


async def test_get_multiple_primary_key(engine):
    u1 = await test_create(engine)
    u2 = await test_create(engine)
    await Friendship.create(bind=engine, my_id=u1.id, friend_id=u2.id)
    with pytest.raises(ValueError,
                       match='Incorrect number of values as primary key'):
        await Friendship.get((u1.id,), bind=engine)
    with pytest.raises(ValueError,
                       match='Incorrect number of values as primary key'):
        await Friendship.get(u1.id, bind=engine)
    f = await Friendship.get((u1.id, u2.id), bind=engine)
    assert f
    assert f.my_id == u1.id
    assert f.friend_id == u2.id


async def test_multiple_primary_key_order():
    import gino

    db1 = await gino.Gino(PG_URL)

    class NameCard(db1.Model):
        __tablename__ = 'name_cards'

        first_name = db1.Column(db1.Unicode(), primary_key=True)
        last_name = db1.Column(db1.Unicode(), primary_key=True)

    await db1.gino.create_all()

    try:
        await NameCard.create(first_name='first', last_name='last')
        nc = await NameCard.get(('first', 'last'))
        assert nc.first_name == 'first'
        assert nc.last_name == 'last'
        with pytest.raises(ValueError, match='expected 2, got 3'):
            await NameCard.get(dict(a=1, first_name='first', last_name='last'))
        with pytest.raises(KeyError, match='first_name'):
            await NameCard.get(dict(first='first', last_name='last'))
        nc = await NameCard.get(dict(first_name='first', last_name='last'))
        assert nc.first_name == 'first'
        assert nc.last_name == 'last'
        nc = await NameCard.get({0: 'first', 1: 'last'})
        assert nc.first_name == 'first'
        assert nc.last_name == 'last'
    finally:
        await db1.gino.drop_all()
        await db1.pop_bind().close()

    db2 = await gino.Gino(PG_URL)

    class NameCard(db2.Model):
        __tablename__ = 'name_cards'

        last_name = db2.Column(db2.Unicode(), primary_key=True)
        first_name = db2.Column(db2.Unicode(), primary_key=True)

    await db2.gino.create_all()

    try:
        await NameCard.create(first_name='first', last_name='last')
        nc = await NameCard.get(('last', 'first'))
        assert nc.first_name == 'first'
        assert nc.last_name == 'last'
        nc = await NameCard.get(dict(first_name='first', last_name='last'))
        assert nc.first_name == 'first'
        assert nc.last_name == 'last'
        nc = await NameCard.get({1: 'first', 'last_name': 'last'})
        assert nc.first_name == 'first'
        assert nc.last_name == 'last'
    finally:
        await db2.gino.drop_all()
        await db2.pop_bind().close()


async def test_connection_as_bind(engine):
    async with engine.acquire() as conn:
        await test_get(conn)


async def test_update(engine, random_name):
    u1 = await test_create(engine)
    await u1.update(nickname=random_name).apply(bind=engine, timeout=10)
    u2 = await User.get(u1.id, bind=engine)
    assert u2.nickname == random_name


async def test_update_missing(engine, random_name):
    from gino.exceptions import NoSuchRowError

    u1 = await test_create(engine)
    rq = u1.update(nickname=random_name)
    await u1.delete(bind=engine)
    with pytest.raises(NoSuchRowError):
        await rq.apply(bind=engine, timeout=10)


async def test_update_multiple_primary_key(engine):
    u1 = await test_create(engine)
    u2 = await test_create(engine)
    u3 = await test_create(engine)
    await Friendship.create(bind=engine, my_id=u1.id, friend_id=u2.id)
    f = await Friendship.get((u1.id, u2.id), bind=engine)
    await f.update(my_id=u2.id, friend_id=u3.id).apply(bind=engine)
    f2 = await Friendship.get((u2.id, u3.id), bind=engine)
    assert f2


async def test_delete(engine):
    u1 = await test_create(engine)
    await u1.delete(bind=engine, timeout=10)
    u2 = await User.get(u1.id, bind=engine)
    assert not u2


async def test_delete_bind(bind):
    u1 = await test_create(bind)
    await u1.delete(timeout=10)
    u2 = await User.get(u1.id)
    assert not u2


async def test_delete_multiple_primary_key(engine):
    u1 = await test_create(engine)
    u2 = await test_create(engine)
    f = await Friendship.create(bind=engine, my_id=u1.id, friend_id=u2.id)
    await f.delete(bind=engine)
    f2 = await Friendship.get((u1.id, u2.id), bind=engine)
    assert not f2


async def test_string_primary_key(engine):
    relations = ['Colleagues', 'Friends', 'Lovers']
    for r in relations:
        await Relation.create(bind=engine, timeout=10, name=r)
    r1 = await Relation.get(relations[0], bind=engine, timeout=10)
    assert r1.name == relations[0]


async def test_lookup_287(bind):
    from gino.exceptions import NoSuchRowError

    class Game(db.Model):
        __tablename__ = 'games'
        game_id = db.Column(db.String(32), unique=True)
        channel_id = db.Column(db.String(1), default='A')

    await Game.gino.create()
    try:
        game_1 = await Game.create(game_id='1', channel_id='X')
        game_2 = await Game.create(game_id='2', channel_id='Y')

        # ordinary update should be fine
        uq = game_1.update(game_id='3')

        with pytest.raises(TypeError,
                           match='Model Game has no table, primary key'):
            # but applying the updates to DB should fail
            await uq.apply()

        with pytest.raises(LookupError,
                           match='Instance-level CRUD operations not allowed'):
            await game_2.delete()
        with pytest.raises(LookupError,
                           match='Instance-level CRUD operations not allowed'):
            await game_2.query.gino.all()
        with pytest.raises(LookupError,
                           match='Instance-level CRUD operations not allowed'):
            await game_2.select('game_id')

        # previous ordinary update still in effect
        assert game_1.game_id == '3'

        assert await Game.select('game_id').gino.all() == [('1',), ('2',)]

        Game.lookup = lambda self: Game.game_id == self.game_id
        with pytest.raises(NoSuchRowError):
            await game_1.update(channel_id='Z').apply()
        await game_2.update(channel_id='Z').apply()
        assert await Game.select('channel_id').gino.all() == [('X',), ('Z',)]
    finally:
        await Game.gino.drop()



import os
import enum
import random
import string
from datetime import datetime

import pytest

from gino import Gino
from gino.dialects.asyncpg import JSONB

DB_ARGS = dict(
    host=os.getenv('DB_HOST', 'localhost'),
    port=os.getenv('DB_PORT', 5432),
    user=os.getenv('DB_USER', 'postgres'),
    password=os.getenv('DB_PASS', ''),
    database=os.getenv('DB_NAME', 'postgres'),
)
PG_URL = 'postgresql://{user}:{password}@{host}:{port}/{database}'.format(
    **DB_ARGS)
db = Gino()


@pytest.fixture
def random_name(length=8) -> str:
    return _random_name(length)


def _random_name(length=8):
    return ''.join(random.choice(string.ascii_letters) for _ in range(length))


class UserType(enum.Enum):
    USER = 'USER'


class User(db.Model):
    __tablename__ = 'gino_users'

    id = db.Column(db.BigInteger(), primary_key=True)
    nickname = db.Column('name', db.Unicode(), default=_random_name)
    profile = db.Column('props', JSONB(), nullable=False, server_default='{}')
    type = db.Column(
        db.Enum(UserType),
        nullable=False,
        default=UserType.USER,
    )
    realname = db.StringProperty()
    age = db.IntegerProperty(default=18)
    balance = db.IntegerProperty(default=0)
    birthday = db.DateTimeProperty(
        default=lambda i: datetime.utcfromtimestamp(0))
    team_id = db.Column(db.ForeignKey('gino_teams.id'))

    @balance.after_get
    def balance(self, val):
        if val is None:
            return 0.0
        return float(val)

    def __repr__(self):
        return '{}<{}>'.format(self.nickname, self.id)


class Friendship(db.Model):
    __tablename__ = 'gino_friendship'

    my_id = db.Column(db.BigInteger(), primary_key=True)
    friend_id = db.Column(db.BigInteger(), primary_key=True)

    def __repr__(self):
        return 'Friends<{}, {}>'.format(self.my_id, self.friend_id)


class Relation(db.Model):
    __tablename__ = 'gino_relation'

    name = db.Column(db.Text(), primary_key=True)


class Team(db.Model):
    __tablename__ = 'gino_teams'

    id = db.Column(db.BigInteger(), primary_key=True)
    name = db.Column(db.Unicode(), default=_random_name)
    parent_id = db.Column(db.ForeignKey('gino_teams.id'))
    company_id = db.Column(db.ForeignKey('gino_companies.id'))

    def __init__(self, **kw):
        super().__init__(**kw)
        self._members = set()

    @property
    def members(self):
        return self._members

    @members.setter
    def add_member(self, user):
        self._members.add(user)


class Company(db.Model):
    __tablename__ = 'gino_companies'

    id = db.Column(db.BigInteger(), primary_key=True)
    name = db.Column(db.Unicode(), default=_random_name)
    logo = db.Column(db.LargeBinary())

    def __init__(self, **kw):
        super().__init__(**kw)
        self._teams = set()

    @property
    def teams(self):
        return self._teams

    @teams.setter
    def add_team(self, team):
        self._teams.add(team)


class UserSetting(db.Model):
    __tablename__ = 'gino_user_settings'

    # No constraints defined on columns
    id = db.Column(db.BigInteger())
    user_id = db.Column(db.BigInteger())
    setting = db.Column(db.Text())
    value = db.Column(db.Text())
    col1 = db.Column(db.Integer, default=1)
    col2 = db.Column(db.Integer, default=2)

    # Define indexes and constraints inline
    id_pkey = db.PrimaryKeyConstraint('id')
    user_id_fk = db.ForeignKeyConstraint(['user_id'], ['gino_users.id'])
    user_id_setting_unique = db.UniqueConstraint('user_id', 'setting')
    col1_check = db.CheckConstraint('col1 >= 1 AND col1 <= 5')
    col2_idx = db.Index('col2_idx', 'col2')


def qsize(engine):
    # noinspection PyProtectedMember
    return engine.raw_pool._queue.qsize()
